#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/longhorn_helper.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/disaster_recovery_utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

FAILED_PVC_LIST=""
export PROFILE=${PROFILE:-""}

########
# Delete PVC Resources
# Input:
# PVC_NAME
# PVC_NAMESPACE
# PV_NAME
# VOLUME_ATTACHMENT_NAME
########
function delete_pvc_resources(){
  local pv_name=$1
  local volumeattachment_name=$2

  info "deleting pv & volumes forcefully"
  delete_pv_forcefully "${pv_name}"
  sleep 2
  delete_longhornvolume_forcefully "${pv_name}"
  sleep 2
  if [ -n "${volumeattachment_name}" ]; then
    info "deleting volumeattachments forcefully"
    delete_volumeattachment_forcefully "${volumeattachment_name}"
    sleep 5
  fi
}

########
# Store PVC Resources
# Input:
# PVC_NAME
# PVC_NAMESPACE
# PV_NAME
# VOLUME_ATTACHMENT_NAME
########
function store_pvc_resources(){
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4

  if [[ -n "${volumeattachment_name}" && "${volumeattachment_name}" != " " ]]; then
    result=$(kubectl get volumeattachments "${volumeattachment_name}" -o json > "${LOCAL_RESTORE_PATH}/${volumeattachment_name}".json) || true
    info "${result}"
  fi

  kubectl get pv "${pv_name}" -o json > "${LOCAL_RESTORE_PATH}/${pv_name}"-pv.json
  kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}" -o json > "${LOCAL_RESTORE_PATH}/${pv_name}"-volume.json
  kubectl get pvc "${pvc_name}" -n "${namespace}" -o json > "${LOCAL_RESTORE_PATH}/${pvc_name}"-pvc.json
}

function wait_pvc_bound() {
  local namespace=$1
  local pvc_name=$2

  try=0
  maxtry=30
  while (( try < maxtry )); do
    local status=""
    status=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".status | select( has(\"phase\")).phase")
    if [[ "${status}"  == "Bound" ]]; then
      info "PVC ${pvc_name} Bouned successfully."
      break;
    fi
    info "waiting for PVC to bound...${try}/${maxtry}"; sleep 30
    try=$((try+1))
  done
}

########
# Store PVC Resources
# Input:
# PVC_NAME
# PVC_NAMESPACE
# PV_NAME
# VOLUME_ATTACHMENT_NAME
########
function create_pvc_resources(){
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  local pvc_uid=""

  kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-volume.json
  kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-pv.json

  pvc_uid=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".metadata.uid")

  # update pv with pvc uid
  kubectl patch pv "${pv_name}" --type json -p "[{\"op\": \"add\", \"path\": \"/spec/claimRef/uid\", \"value\": \"${pvc_uid}\"}]"

  wait_pvc_bound "${namespace}" "${pvc_name}"
  info "${result}"
}

function create_volume() {
  local pv_name=$1
  local backup_path=$2
  local response
  create_volume_status="succeed"
  info "Creating Volume with PV: ${pv_name} and BackupPath: ${backup_path}"
  local accessmode
  local replicacount
  accessmode=$(< "${LOCAL_RESTORE_PATH}/${pv_name}-volume.json" jq -r ".spec.accessMode")
  replicacount=$(< "${LOCAL_RESTORE_PATH}/${pv_name}-volume.json" jq -r ".spec.numberOfReplicas")

  # create volume from backup pv
  response=$(curl --noproxy "*" "${LONGHORN_URL}/v1/volumes" -H 'Accept: application/json' -H 'Content-Type: application/json;charset=UTF-8' --data-raw "{\"name\":\"${pv_name}\", \"accessMode\":\"${accessmode}\", \"fromBackup\": \"${backup_path}\", \"numberOfReplicas\": ${replicacount}}")

  sleep 5

  if [[ -n "${response}" && "${response}" != "null" && "${response}" != " " && -n $(echo "${response}"|jq -r ".id") ]]; then
  # wait for volume to be detached , max wait 4hr
    local try=0
    local maxtry=480
    local success=0
    while (( try < maxtry ));do
      status=$(kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}"  -o json |jq -r ".status.state") || true
      info "volume ${pv_name} status: ${status}"
      if [[ "${status}" == "detached" ]]; then
        # update label
        kubectl -n longhorn-system label volumes.longhorn.io/"${pv_name}" recurring-job-group.longhorn.io/uipath-backup=enabled
        success=1
        info "Volume ready to use"
        break
      fi
      info "waiting for volume to be ready to use${try}/${maxtry}..."; sleep 30
      restore_status_url="${LONGHORN_URL}/v1/volumes/${pv_name}?action=snapshotList"
      restore_progress "${restore_status_url}"
      try=$(( try + 1 ))
    done

    if [ "${success}" -eq 0 ]; then
      create_volume_status="failed"
      error_without_exit "${pv_name} Volume is not ready to use with status ${status}"
    fi
  else
    create_volume_status="failed"
    kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-volume.json || true
    error_without_exit "${pv_name} Volume creation failed ${response} "
  fi
  echo "${create_volume_status}"
}

function restore_withoutbackupvolume() {
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  store_pvc_resources "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  sleep 5
  delete_pvc_resources "${pv_name}" "${volumeattachment_name}"
  sleep 5
  create_pvc_resources "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  sleep 5
}


function enable_disable_self_heal_for_argocd_app() {
  local application="$1"
  local input_op="$2"
  local op
  case "$input_op" in
    enable)
      op=true
      ;;
    disable)
      op=false
      ;;
    *)
      error "Unsupported option"
      ;;
  esac
  kubectl -n argocd patch application "${application}" --type=json -p "[
    {\"op\":\"replace\",\"path\":\"/spec/syncPolicy/automated/selfHeal\",\"value\":${op}}]" || true
}

function cleanup_ceph_resources(){
  local namespace="rook-ceph"
  local k8s_resources_to_clean=(job pod serviceaccount service secret configmap persistentvolumeclaim)

  for resource in "${k8s_resources_to_clean[@]}"; do
    kubectl -n "${namespace}"  get "${resource}" --no-headers -o name  | grep "${resource}" | xargs -r kubectl -n "${namespace}"  patch --type=json -p '[{"op":"remove" , "path": "/metadata/finalizers"}]' || true
    kubectl -n "${namespace}"  get "${resource}" --no-headers -o name  | grep "${resource}" | xargs -r kubectl -n "${namespace}" delete --force
  done
}

function restore_with_backupvolume() {
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  local backup_path=$5
  create_volume_status="succeed"
  store_pvc_resources "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  sleep 5
  delete_pvc_resources "${pv_name}" "${volumeattachment_name}"
  sleep 5
  create_volume "${pv_name}" "${backup_path}"
  sleep 10

  kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-pv.json
  if [[ -n "${create_volume_status}" && "${create_volume_status}" != "succeed" ]]; then
    error_without_exit "Backup volume restore failed for pvc ${pvc_name} in namespace ${namespace}"
    restore_pvc_status="failed"
  else 
    local pvc_uid=""
    pvc_uid=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".metadata.uid")

    # update pv with pvc uid
    kubectl patch pv "${pv_name}" --type json -p "[{\"op\": \"add\", \"path\": \"/spec/claimRef/uid\", \"value\": \"${pvc_uid}\"}]"

    wait_pvc_bound "${namespace}" "${pvc_name}"
    info "${result}"
  fi
  echo ${restore_pvc_status}
}

function restore_pvc(){
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  local backup_path=$5
  local backup_pvc=${6:-true}
  restore_pvc_status="succeed"
  if [[ -n "${backup_path}" && "${backup_pvc}" == "true" ]]; then
    restore_with_backupvolume "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}"
  else
    restore_withoutbackupvolume "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  fi
  echo ${restore_pvc_status}
  # attach_volume "${pv_name}"
}

function attach_volume() {
  local pv_name=$1
  local node_name

  # attach volume to node
  node_name=$(curl --noproxy "*" "${LONGHORN_URL}/v1/nodes?"|jq -r ".data[0].name")
  curl --noproxy "*" "${LONGHORN_URL}/v1/volumes/${pv_name}?action=attach" --data-raw "{\"hostId\":\"${node_name}\"}"
}

function get_pvc_resources() {

  get_pvc_resources_resp=""

  local PVC_NAME=$1
  local PVC_NAMESPACE=$2

  # PV have one to one mapping with PVC
  PV_NAME=$(kubectl -n "${PVC_NAMESPACE}" get pvc "${PVC_NAME}" -o json|jq -r ".spec.volumeName")

  VOLUME_ATTACHMENT_LIST=$(kubectl get volumeattachments -n "${PVC_NAMESPACE}" -o=json|jq -c ".items[]|\
{name: .metadata.name, pvClaimName:.spec.source| select( has (\"persistentVolumeName\")).persistentVolumeName}")

  VOLUME_ATTACHMENT_NAME=""
  for VOLUME_ATTACHMENT in ${VOLUME_ATTACHMENT_LIST};
  do
    PV_CLAIM_NAME=$(echo "${VOLUME_ATTACHMENT}"|jq -r ".pvClaimName")
    if [[ "${PV_NAME}" = "${PV_CLAIM_NAME}" ]]; then
      VOLUME_ATTACHMENT_NAME=$(echo "${VOLUME_ATTACHMENT}"|jq -r ".name")
      break;
    fi
  done

  BACKUP_PATH=""
  validate_pv_backup_available "${PV_NAME}"
  local is_backup_available=${IS_BACKUP_AVAILABLE:- 0}
  unset IS_BACKUP_AVAILABLE

  if [ "${is_backup_available}" -eq 1 ]; then
    info "Backup is available for PVC ${PVC_NAME}"

    get_backup_list_by_pvname "${PV_NAME}"
    BACKUP_PATH=$(echo "${PV_BACKUP_PATH}"| jq -r ".url")
    unset PV_BACKUP_PATH
  fi

  get_pvc_resources_resp="{\"pvc_name\": \"${PVC_NAME}\", \"pv_name\": \"${PV_NAME}\", \"volumeattachment_name\": \"${VOLUME_ATTACHMENT_NAME}\", \"backup_path\": \"${BACKUP_PATH}\"}"
  echo "${get_pvc_resources_resp}"
}

function restore_pvc_attached_to_deployments() {
  local namespace="${1}"
  local deployment_list
  # list of all deployment using PVC
  deployment_list=$(kubectl get deployments -n "${namespace}" -o=json | jq -c ".items[] |\
{name: .metadata.name, namespace: .metadata.namespace, replicas: .spec.replicas, pvcList:.spec.template.spec |\
[select( has (\"volumes\") ).volumes[] |\
select( has (\"persistentVolumeClaim\") ).persistentVolumeClaim.claimName]|select(length > 0) }")

  for deployment in ${deployment_list};
  do
    local pvc_restore_failed
    pvc_restore_failed=""
    deployment_name=$(echo "${deployment}"| jq -r ".name")
    pvc_list=$(echo "${deployment}"| jq -r ".pvcList[]")
    namespace=$(echo "${deployment}"| jq -r ".namespace")
    replica=$(echo "${deployment}" | jq -r ".replicas")
    restore_pvc_status=""
    scale_down_deployment "${deployment_name}" "${namespace}"

    for pvc_name in ${pvc_list};
    do
      local pvc_exist
      pvc_exist=$(kubectl -n "${namespace}" get pvc "${pvc_name}") || true
      if [[ -z "${pvc_exist}" || "${pvc_exist}" == " " ]]; then
        warn "PVC not available for the deployment ${deployment}, skipping restore."
        continue;
      fi
      local pvc_storageclass
      pvc_storageclass=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".spec.storageClassName")
      if [[ ! ( "${pvc_storageclass}" == "${STORAGE_CLASS}" || "${pvc_storageclass}" == "${STORAGE_CLASS_SINGLE_REPLICA}" ) ]]; then
        info "backup not available for pvc ${pvc_name}, storageclass: ${pvc_storageclass} "
        continue;
      fi

      # get pv, volumeattachments for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      local volumeattachment_name
      local backup_path

      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")
      volumeattachment_name=$(echo "${get_pvc_resources_resp}"| jq -r ".volumeattachment_name")
      backup_path=$(echo "${get_pvc_resources_resp}"| jq -r ".backup_path")

      if [[ -z "${backup_path}" || "${backup_path}" == " " || "${backup_path}" == "null" ]]; then
        error_without_exit "pvc backup path for deployment ${deployment_name} in namespace ${namespace} is empty"
        pvc_restore_failed="error"
        FAILED_PVC_LIST="${FAILED_PVC_LIST},${pv_name}"
        continue;
      fi
      restore_pvc_status="succeed"
      restore_pvc "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}"

      if [[ -n "${restore_pvc_status}" && "${restore_pvc_status}" != "succeed" ]]; then
        pvc_restore_failed="error"
        FAILED_PVC_LIST="${FAILED_PVC_LIST},${pv_name}"
        continue;
      fi
    done

    sleep 10
    scale_up_deployment "${deployment_name}" "${namespace}" "${replica}"
    sleep 5

    if [[ -n "${pvc_restore_failed}" && "${pvc_restore_failed}" == "error" ]]; then
        error_without_exit "Failed to restore pvc for deployment ${deployment_name} in namespace ${namespace}."
    fi
    
  done
}

function restore_pvc_attached_to_statefulsets() {
  local namespace
  local backup_pvc
  namespace="${1}"
  backup_pvc=${2:-true}
  local statefulset_list

  # list of all statefulset using PVC
  statefulset_list=$(kubectl get statefulset -n "${namespace}" -o=json | jq -r ".items[] | select(.spec.volumeClaimTemplates).metadata.name")

  for statefulset_name in ${statefulset_list};
  do
    local replica
    local ownerReferences
    local pvc_restore_failed
    local try=0
    local maxtry=5
    local status="notready"
    pvc_restore_failed=""
    restore_pvc_status=""

    # check if statefulset is reday
    while [[ "${status}" == "notready" ]] && (( try < maxtry )); do
      info "fetch statefulset ${statefulset_name} metadata...  ${try}/${maxtry}"
      try=$(( try + 1 ))
      replica=$(kubectl get statefulset "${statefulset_name}" -n "${namespace}" -o=json | jq -c ".spec.replicas")

      if [[ "${replica}" != 0 ]]; then
        status="ready"
      else
        info "statefulset ${statefulset_name} replica is not ready. Wait and retry"; sleep 30
      fi
    done

    if [[ "${status}" != "ready" ]]; then
      error_without_exit "Failed to restore pvc for Statefulset ${statefulset_name} in namespace ${namespace}. Please retrigger volume restore step."
    fi

    # Fetch ownerReferences and claim name
    ownerReferences=$(kubectl get statefulset "${statefulset_name}" -n "${namespace}" -o=json | jq -c ".metadata.ownerReferences")
    claimTemplatesName=$(kubectl get statefulset "${statefulset_name}" -n "${namespace}" -o=json | jq -c ".spec | select( has (\"volumeClaimTemplates\") ).volumeClaimTemplates[].metadata.name " | xargs)
    
    info "Scaling down Statefulset ${statefulset_name} with ${replica} under namespace ${namespace}"
    scale_down_statefulset "${statefulset_name}" "${namespace}" "${ownerReferences}"
    for name in ${claimTemplatesName}; do
      local pvc_prefix
      pvc_prefix="${name}-${statefulset_name}"

      for((i=0;i<"${replica}";i++)); do
        local pvc_name
        pvc_name="${pvc_prefix}-${i}"

        pvc_exist=$(kubectl -n "${namespace}" get pvc "${pvc_name}") || true
        if [[ -z "${pvc_exist}" || "${pvc_exist}" == " " ]]; then
          warn "PVC not available for the statefulset ${statefulset_name}, skipping restore."
          continue;
        fi

        local pvc_storageclass
        pvc_storageclass=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".spec.storageClassName")
        if [[ ! ( "${pvc_storageclass}" == "${STORAGE_CLASS}" || "${pvc_storageclass}" == "${STORAGE_CLASS_SINGLE_REPLICA}" ) ]]; then
          info "backup not available for pvc ${pvc_name}, storageclass: ${pvc_storageclass} "
          continue;
        fi

        # get pv, volumeattachments for pvc
        get_pvc_resources_resp=""
        get_pvc_resources "${pvc_name}" "${namespace}"

        local pv_name
        local volumeattachment_name
        local backup_path
        pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")
        volumeattachment_name=$(echo "${get_pvc_resources_resp}"| jq -r ".volumeattachment_name")
        backup_path=$(echo "${get_pvc_resources_resp}"| jq -r ".backup_path")
        if [[ -z "${backup_path}" || "${backup_path}" == " " || "${backup_path}" == "null" ]]; then
          pvc_restore_failed="error"
          FAILED_PVC_LIST="${FAILED_PVC_LIST},${pv_name}"
          continue;
        fi

        restore_pvc_status="succeed"
        restore_pvc "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}" "${backup_pvc}"
        if [[ -n "${restore_pvc_status}" && "${restore_pvc_status}" != "succeed" ]]; then
          pvc_restore_failed="error"
          FAILED_PVC_LIST="${FAILED_PVC_LIST},${pv_name}"
          continue;
        fi
      done
    done

    sleep 10
    scale_up_statefulset "${statefulset_name}" "${namespace}" "${replica}" "${ownerReferences}"
    sleep 5

    if [[ -n "${pvc_restore_failed}" && "${pvc_restore_failed}" == "error" ]]; then
        error_without_exit "Failed to restore pvc for Statefulset ${statefulset_name} in namespace ${namespace}."
    fi    
  done
}

# rook-ceph disaster-recovery https://github.com/rook/rook/blob/master/Documentation/ceph-disaster-recovery.md
function restore_rook_ceph_cluster() {

  local namespace="rook-ceph"
  local objectstore_name="rook-ceph"

  # Disable self heal
  enable_disable_self_heal_for_argocd_app "fabric-installer" "disable"
  enable_disable_self_heal_for_argocd_app "rook-ceph-operator" "disable"
  enable_disable_self_heal_for_argocd_app "rook-ceph-object-store" "disable"
  # scale down operator deployment
  scale_down_deployment "rook-ceph-operator" "${namespace}"

  local deployment_list
  # delete all deployment except "rook-ceph-operator" & "rook-ceph-tools"
  deployment_list=$(kubectl -n "${namespace}" get deploy -o json| jq -r ".items[].metadata.name")
  for deployment in ${deployment_list};
  do
    if [[ "${deployment}" != *"rook-ceph-operator"* && "${deployment}" != *"rook-ceph-tools"* ]]; then
      kubectl -n "${namespace}" delete deploy "${deployment}"
    fi
  done

  cleanup_ceph_resources
  for dsPod in $(kubectl -n kube-system get pod -l app=longhorn-loop-device-cleaner -o name); do
    kubectl -n kube-system exec "${dsPod}"  -- rm -rf /var/lib/rook
  done

  # Enable self heal
  enable_disable_self_heal_for_argocd_app "fabric-installer" "enable"
  enable_disable_self_heal_for_argocd_app "rook-ceph-operator" "enable"
  enable_disable_self_heal_for_argocd_app "rook-ceph-object-store" "enable" 

  scale_up_deployment "rook-ceph-operator" "${namespace}" "1"

  if wait_till_cephobjectstore_ready "${objectstore_name}" "${namespace}" "app=rook-ceph-rgw"; then
    kubectl -n "${namespace}" rollout restart  deploy/rook-ceph-tools
    kubectl -n "${namespace}" rollout status deploy/rook-ceph-tools
    argocd app sync fabric-configure >> /dev/null 2>&1
    kubectl -n "${namespace}" get secret ceph-object-store-secret --show-labels  | grep -q 'config-discovery=yes' || error "Secret: 'ceph-object-store-secret' does not exists in namespace: '${namespace}'"
    kubectl -n uipath-infra get job "credential-manager-job" -o json | jq 'del(. | .spec.selector, .spec.template.metadata.labels)' | kubectl replace --force -f -
    kubectl -n uipath-infra wait --timeout=10m --for=condition=complete "job/credential-manager-job"
    #shellcheck disable=SC2119
    restore_ceph && helm -n uipath-infra uninstall backup-service
  fi
}

function restore_fabric_component() {

  # exclude namespace for component which are handled explicitly
  local namespace_list
  namespace_list=$(kubectl get namespace -o json|jq -r ".items[].metadata| \
select( .name != \"docker-registry\" \
and .name != \"uipath\" \
and .name != \"rook-ceph\" \
and .name != \"redis-system\").name")

  for namespace in ${namespace_list}; do
    restore_pvc_attached_to_deployments "${namespace}"
    restore_pvc_attached_to_statefulsets "${namespace}"
  done
restore_rook_ceph_cluster
kubectl -n cattle-logging-system get pod -l statefulset.kubernetes.io/pod-name | grep 'CrashLoopBackOff' | awk '{print $1}' | xargs -r kubectl -n cattle-logging-system delete pod

}

#shellcheck disable=SC2120
function restore_ceph(){
  # If you want to restore a specific bucket use below config
  local include_buckets="${1:-""}"           ## comma separated list of buckets to restore
  local include_buckets_pattern="${2:-""}"   ## unanchored perl regex pattern of buckets to restore
  local exclude_buckets="${3:-""}"           ## comma separated list of buckets to not restore 
  local exclude_buckets_pattern="${4:-""}"  ## unanchored perl regex pattern of buckets to not restore
  local restore_epoch="${5:-latest}"       ## Restore buckets to specific point in time using epoch value
  local restore_env=[]
  local epoch

  [[ -n "${include_buckets}" ]] && restore_env=$(echo "$restore_env" | jq --arg key "INCLUDE_BUCKETS" --arg value "$include_buckets" '. += [{"name": $key, "value": $value}]')
  [[ -n "${include_buckets_pattern}" ]] && restore_env=$(echo "$restore_env" | jq --arg key "INCLUDE_BUCKETS_PATTERN" --arg value "$include_buckets_pattern" '. += [{"name": $key, "value": $value}]')
  [[ -n "${exclude_buckets}" ]] && restore_env=$(echo "$restore_env" | jq --arg key "EXCLUDE_BUCKETS" --arg value "$exclude_buckets" '. += [{"name": $key, "value": $value}]')
  [[ -n "${exclude_buckets_pattern}" ]] && restore_env=$(echo "$restore_env" | jq --arg key "EXCLUDE_BUCKETS_PATTERN" --arg value "$exclude_buckets_pattern" '. += [{"name": $key, "value": $value}]')
  [[ -n "${restore_epoch}" && $restore_epoch != "latest" ]] && restore_env=$(echo "$restore_env" | jq --arg key "RESTORE_EPOCH" --arg value "$restore_epoch" '. += [{"name": $key, "value": $value}]')

  epoch=$(date +"%s")
  kubectl -n uipath-infra get cronjob backup-ceph-objectstore-cronjob >> /dev/null 2>&1 || error "Unable to find ceph backup cronjob"
  kubectl -n uipath-infra get cronjob backup-ceph-objectstore-cronjob -o json  | jq --argjson restoreEnv "${restore_env}" '
    del(.metadata.annotations, .metadata.creationTimestamp, .metadata.labels, .metadata.resourceVersion, .metadata.uid, .status) | .spec.suspend = true | .metadata.name = "restore-ceph-objectstore-cronjob" | 
    .spec.jobTemplate.spec.template.spec.containers[0].name = "ceph-objectstore-restore" | (.spec.jobTemplate.spec.template.spec.containers[0].env[] | select(.name=="OP").value) |= "restore" | .spec.jobTemplate.spec.template.spec.containers[0].env += $restoreEnv' | kubectl apply -f -
  kubectl -n uipath-infra create job "restore-ceph-objectstore-cronjob-${epoch}"  --from cj/restore-ceph-objectstore-cronjob

  local job_status=failed
  while true; do
    kubectl -n "uipath-infra" wait --timeout=10s --for=condition=complete "job/restore-ceph-objectstore-cronjob-${epoch}" && job_status=passed && break
    kubectl -n "uipath-infra" wait --timeout=10s --for=condition=Failed "job/restore-ceph-objectstore-cronjob-${epoch}" && break
  done
  kubectl -n "uipath-infra" logs "job/restore-ceph-objectstore-cronjob-${epoch}"
  if [[ $job_status == "passed" ]]; then
    kubectl -n uipath-infra delete cj/restore-ceph-objectstore-cronjob
  else
    echo "Ceph restore job ${job_status}"
    return 1
  fi
}

function restore_services_component() {
  namespace_list="uipath"
  for namespace in ${namespace_list}; do
    restore_pvc_attached_to_deployments "${namespace}"
    restore_pvc_attached_to_statefulsets "${namespace}"
  done
  # clean any pod with corrupt io error
  clean_pod_with_io_error "uipath"

  # rollout deployments in uipath namespace
  kubectl rollout restart deployment --namespace "uipath"
  if wait_for_all_pods_in_namespace_ready "uipath"; then
    info "Rollback successful"
  else
    error_without_exit "Rollback of deployments failed."
  fi
}

function restore_infra_component() {
  namespace_list="docker-registry"
  for namespace in ${namespace_list}; do
    restore_pvc_attached_to_deployments "${namespace}"
    restore_pvc_attached_to_statefulsets "${namespace}"
  done
}

########
# Restore PVC & Deployments
########
function restore_pvcs_and_dependent_deployments() {
  if [[ "${PROFILE}" != "" ]]; then
    # validate profile
    check_available_masters
  fi

  set_longhorn_url
  validate_longhorn_reachable; sleep 120
  validate_longhorn_pods_health

  # update settings for backup
  update_default_settings "default-longhorn-static-storage-class" "${STORAGE_CLASS}"
  update_default_settings "backup-target" "nfs://${NFS_ENDPOINT}:${NFS_MOUNT_PATH}"

  # validate longhorn backup reachable
  validate_longhorn_backup_reachable; sleep 120

  # Save restored data in case of failures
  LOCAL_RESTORE_PATH="restoredata"
  mkdir -p ${LOCAL_RESTORE_PATH}
  export LOCAL_RESTORE_PATH="${LOCAL_RESTORE_PATH}"

  restore_infra_component

  # wait for docker registry to come up
  if [[ "${INSTALL_TYPE}" == "offline" ]]; then
    info "Waiting for all pods to pick images from private docker registry..."
    imagepullbackoffcount=-1
    try=0
    maxtry=120
    while(( try < maxtry )); do
      # kubectl get pods -A | grep -v "Running"| grep -v "Complete"| grep -v "Terminating"; sleep 10
      try=$(( try + 1 ))
      imagepullbackoffcount=$(kubectl get pods -A | grep -c "ImagePullBackOff")
      info "Total pods in ImagePullBackOff ${imagepullbackoffcount}, Waiting for all pods to come up... ${try}/${maxtry}"; sleep 30
      if [[ "${imagepullbackoffcount}" -lt 1 ]]; then
        break
      fi
    done

    if [[ "${imagepullbackoffcount}" -ge 1 ]]; then
      error "Failed to restore images for services"
    fi
  fi

  # check argocd is up
  try=0
  maxtry=10
  ready=0
  while(( try < maxtry )); do
    is_argocd_up
    ready="$?"
    if (( ready == 1 )) ; then
      break;
    fi
    info "Waiting for argocd to come up...  ${try}/${maxtry}"; sleep 20
    try=$(( try + 1 ))
  done

  if (( ready == 0 )); then
    error "Failed to start argocd."
  fi

  get_fabric_install_paths "${SCRIPT_DIR}"
  get_argocd_install_params
  ARGO_FQDN="$(get_mandatory_json_value ".subdomain.argo_cd").$(get_mandatory_json_value ".fqdn")"

  login_argocd "${ARGOCD_NODEPORT_HTTP}" "${ARGO_FQDN}" "${ARGOCD_PASSWORD_SECRET_NAME}" "${ARGOCD_NAMESPACE}" "argocd"

  restore_fabric_component
  restore_services_component

  if [[ -n "${FAILED_PVC_LIST}" ]]; then
    error "Failed to restore PVC List: ${FAILED_PVC_LIST}"
  else
    rm -rf ${LOCAL_RESTORE_PATH}
    info "PVC and Deployment successfully restored..."
  fi
}
