#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MESSAGE_FILE="${SCRIPT_DIR}/../readiness_message_en.json"
if [[ ! -f "${MESSAGE_FILE}" ]] ; then
  echo "Certificate Readiness message file not present"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

##deployment_type online|offline. Default to online 
DEPLOYMENT_TYPE="online"
##Comma separated list of node's private/internal ips passed via argument
NODES=""
##Config file PATH
CONFIG_FILE_NAME="config.json"
CONFIG="${SCRIPT_DIR}/${CONFIG_FILE_NAME}"
##enabled services
ENABLED_SERVICES_JSON_ARRAY='[]'
##exit code
EXIT_CODE=0
## DNS nameservers
DNS_NAMESERVERS=()
## DNS TCP
IS_DNS_TCP_ENABLED="false"
## DNS search domains
DNS_SEARCH_DOMAINS=""

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Run connectivity checks between nodes, outbound connection test for online setup"
  echo
  echo "Arguments"
  echo "  -m|--machines                      Comma separated IP address / hostnames of all the machines to run connectivity check against"
  echo "  --install-type                     Possible values will be online or offline. Defaults to online"
  echo "  -s|--enabled-services              Comma separated list of enabled services"
  echo "  -d|--debug                         Increase logging verbosity to show all debug logs"
  echo "  -h|--help                          Show this help message and exit"
  echo  
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -m|--machines)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODES="$2"
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;     

    -s|--enabled-services)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        ENABLED_SERVICES="$2"
        ENABLED_SERVICES_JSON_ARRAY=$(jq -c -n --arg services "${ENABLED_SERVICES}" '[$services | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;   

    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        DEPLOYMENT_TYPE="$2"
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;

    -*|*) # unsupported flags
      # display_usage
      # echo "Error: Unsupported flag $1"
      # exit 1
      shift
      ;;
    esac
  done

  # [[ -z $NODES ]] && echo "Error: Required argument -m|--machines missing while running connectivity check" && exit 1
  NODES_JSON_ARRAY=$(jq -c -n --arg nodes "${NODES}" '[$nodes | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
  [[ ! -f ${CONFIG} ]] && echo "Error: config file '${CONFIG}' not found" && exit 1
  
}

parse_prereq_input_file() {
    CUSTOM_DNS_RESOLVER="$(get_optional_json_value ".infra.custom_dns_resolver")"
}

function transport_connectivity_check() {
  local proto="$1"
  local ports="$2"
  local extra_arg=""
  local node_array
  local timeout=5
  local output
  local connection_status
  local node
  local port
  
  [[ $proto == "udp" ]] &&  extra_arg="-u"
  node_array=$(echo "$NODES_JSON_ARRAY" | jq -r '.[]')
  while IFS=$'\n' read -r node
  do
    node=$(printf '%s' "$node" | tr -d '[:space:]')
    [[ -z "${node}" ]] && continue
    while IFS=$'\n' read -r port
    do
      port=$(printf '%s' "$port" | tr -d '[:space:]')
      [[ -z "${port}" ]] && continue
      output=$(nc -z -v ${extra_arg} -w ${timeout} "${node}"  "${port}" 2>&1)
      case "$output" in
        *"bytes sent"*|*"succeeded"*)
          connection_status="connected"
          ;;
        *"Connection refused"*)
          connection_status="refused"
          ;;
        *"timed out"*)
          connection_status="timeout"
          ;;
        *"unreachable"*)
          connection_status="unreachable"
          ;;
        "")
          connection_status="empty"
          ;;
        *)
          connection_status="unknown"
          ;;
      esac

      if [[ $proto == "tcp" ]] ; then
        if [[ $connection_status == "connected" ]] || [[ $connection_status == "refused" ]] ; then
          add_test_result "PASSED" "$(get_readiness_message ".connectivity.transport_connectivity_info" "${MESSAGE_FILE}" "${node}" "${proto}" "${port}" )"
        else
          add_test_result "FAILED" "$(get_readiness_message ".connectivity.transport_connectivity_error" "${MESSAGE_FILE}" "${node}" "${proto}" "${port}" )" "$(get_readiness_mitigation ".connectivity.transport_connectivity_error" "${MESSAGE_FILE}" "${port}" )" 
        fi
      else
        if [[ $connection_status == "refused" || $connection_status == "empty" ]] ; then
          add_test_result "PASSED" "$(get_readiness_message ".connectivity.transport_connectivity_info" "${MESSAGE_FILE}" "${node}" "${proto}" "${port}" )"
        elif [[ $connection_status == "connected" ]] ; then
          add_test_result "WARN" "$(get_readiness_message ".connectivity.transport_connectivity_warn" "${MESSAGE_FILE}" "${proto}" "${port}" "${node}" "${node}" "${proto}" "${port}" )" "$(get_readiness_mitigation ".connectivity.transport_connectivity_warn" "${MESSAGE_FILE}" "${port}" )"
        else
          add_test_result "FAILED" "$(get_readiness_message ".connectivity.transport_connectivity_error" "${MESSAGE_FILE}" "${node}" "${proto}" "${port}" )" "$(get_readiness_mitigation ".connectivity.transport_connectivity_error" "${MESSAGE_FILE}" "${port}" )" 
        fi
      fi
    done <<< "${ports}"
  done <<< "${node_array}"
}

function outbound_connectivity_checks() {
  local url_list
  local url
  local output
  local telemetry_opt_out
  telemetry_opt_out=$(get_json_value_or_return_default_value ".telemetry_optout" "false")

  url_list=$(jq -r  --argjson enabledServices "${ENABLED_SERVICES_JSON_ARRAY}" --arg telemetry "${telemetry_opt_out}" '.outbound_checks | to_entries[]  | select((.key as $k | $enabledServices | index($k)) or .key=="common" or ($telemetry=="true" and .key=="telemetry")).value[]' "${CONFIG}")
  while IFS=$'\n' read -r url
  do
    url=$(printf '%s' "$url" | tr -d '[:space:]')
    [[ -z "${url}" ]] && continue
    output=$(timeout 12s curl -m 10  -w "%{http_code}" --silent --output /dev/null "${url}")
    if [[ $output == "000" ]] ; then
      add_test_result "FAILED" "$(get_readiness_message ".connectivity.connectivity_error" "${MESSAGE_FILE}" "${url}" )" "$(get_readiness_mitigation ".connectivity.connectivity_error" "${MESSAGE_FILE}" )"
    elif [[ -z $output ]] ; then
      add_test_result "FAILED" "$(get_readiness_message ".connectivity.connectivity_timeout" "${MESSAGE_FILE}" "${url}" )" "$(get_readiness_mitigation ".connectivity.connectivity_timeout" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "$(get_readiness_message ".connectivity.connectivity_info" "${MESSAGE_FILE}" "${url}" )"
    fi
  done <<< "${url_list}"
}

function dns_resolve() {
    local domain="$1"
    local ips
    local resolved_ip
    local nameserver
    local is_domain_resolved_via_nameserver="false"
    local dig_args=()
    local is_failed="false"
    local dns_search_domain
    local dns_search_domain_ip

    ips=$(timeout 10s getent ahosts  "${domain}")
    case "$?" in
      0)
      ;;
      2)
        add_test_result "FAILED" "$(get_readiness_message ".connectivity.resolution_failure" "${MESSAGE_FILE}" "${domain}")" "$(get_readiness_mitigation ".connectivity.resolution_failure" "${MESSAGE_FILE}")" 
      ;;
      124)
        add_test_result "FAILED" "$(get_readiness_message ".connectivity.resolution_timeout" "${MESSAGE_FILE}" "${domain}")" "$(get_readiness_mitigation ".connectivity.resolution_timeout" "${MESSAGE_FILE}")" 
      ;;
      *)
        add_test_result "FAILED" "$(get_readiness_message ".connectivity.resolution_unexpected_error" "${MESSAGE_FILE}" "${domain}")"
      ;;
    esac
    ips=$(echo "${ips}" | awk '{print $1}' | sort | uniq)
    [[ -z "$ips" ]] && return

    for dns_search_domain in $DNS_SEARCH_DOMAINS
    do
      dns_search_domain_ip=$(timeout 10s getent ahosts "${domain%.}.${dns_search_domain}" | awk '{print $1}' | sort | uniq)
      [[ -n $dns_search_domain_ip ]] && add_test_result "FAILED"   "$(get_readiness_message ".connectivity.resolution_search_domain_failure" "${MESSAGE_FILE}" "${domain}" "${dns_search_domain}")"
    done

    if [[ ${#DNS_NAMESERVERS[@]} != 0 ]] ; then
      [[ "$IS_DNS_TCP_ENABLED" == "true" ]] && dig_args+=(+tcp)
      dig_args+=(+short)
      dig_args+=(+timeout=0)
      for nameserver in "${DNS_NAMESERVERS[@]}"
      do
        resolved_ip=$(dig @"${nameserver}" "${dig_args[@]}" "${domain}" A "${domain}" AAAA 2>&1)
        if [[ $resolved_ip == *"connection timed out"* ]] ; then
          add_test_result "FAILED" "$(get_readiness_message ".connectivity.resolution_nameserver_timeout" "${MESSAGE_FILE}" "${nameserver}")" "$(get_readiness_mitigation ".connectivity.resolution_nameserver_timeout" "${MESSAGE_FILE}")" && is_failed="true" && continue
        fi
        if [[ -z "${resolved_ip}" ]] ; then
          add_test_result "FAILED" "$(get_readiness_message ".connectivity.resolution_nameserver_failure" "${MESSAGE_FILE}" "${nameserver}" "${domain}")" "$(get_readiness_mitigation ".connectivity.resolution_nameserver_failure" "${MESSAGE_FILE}" "${nameserver}")"
          is_failed="true"
        else
          is_domain_resolved_via_nameserver="true"
        fi
      done
      [[ $is_failed == "false" && $is_domain_resolved_via_nameserver == "false" ]] && add_test_result "FAILED" "$(get_readiness_message ".connectivity.resolution_local_error" "${MESSAGE_FILE}" "${domain}")"
    fi
    echo "${ips}"
    return "${EXIT_CODE}"
}

function validate_ip() {
    local ips="$1"
    local domain="$2"
    local msg=""
    local ip_iter
    local exitCode=0

    [[ -z ${domain} ]] || msg=" for ${domain}"
    for ip_iter in $ips
    do
      [[ -z "$ip_iter" ]] && continue
      if ! validate_ipaddress "${ip_iter}" >> /dev/null 2>&1 ; then
        add_test_result "FAILED" "$(get_readiness_message ".connectivity.invalid_ip" "${MESSAGE_FILE}" "${ips//$'\n'/,}" "${msg}" )"
        exitCode=1
        break
      fi
    done
    return "${exitCode}"
}

function validate_subdomains() {
    local fqdn
    local fqdn_ips
    local subdomain_keys=(".subdomain.rook_ceph" ".subdomain.rancher_server" ".subdomain.argo_cd")
    local subdomain_key
    local subdomain_value
    local subdomain
    local subdomain_ips
    local insight_enabled
    local subdomain_resolve_map={}

    parse_fqdn_values

    [[ ${DEPLOYMENT_TYPE} == "offline" ]] && subdomain_keys+=(".subdomain.docker_registry")
    insight_enabled="$(get_mandatory_json_value ".insights.enabled")"
    [[ ${insight_enabled} == "true" ]] && subdomain_keys+=(".subdomain.looker")

    fqdn="${ALTERNATIVE_FQDN}"
    if [[ -z "${fqdn}" ]]; then  
      add_test_result "FAILED" "$(get_readiness_message ".connectivity.fqdn_missing_error" "${MESSAGE_FILE}" )"
    else
      fqdn_ips=$(dns_resolve "${fqdn}")
      EXIT_CODE="$?"
      if [[ -z "${fqdn_ips}" ]]
      then
        add_test_result "FAILED" "$(get_readiness_message ".connectivity.subdomain_ip_missing_error" "${MESSAGE_FILE}" "${fqdn}" )"
      else
        subdomain_resolve_map=$(echo "${subdomain_resolve_map}" |  jq --arg key "${fqdn}" --arg value "$(echo -n "${fqdn_ips}" | tr '\n' ',')" '.[$key]=($value)')
        validate_ip "${fqdn_ips}" "${fqdn}"
      fi
      for index in "${!subdomain_keys[@]}"
      do
          subdomain_key="${subdomain_keys[$index]}"
          subdomain_value="$(get_mandatory_json_value "${subdomain_key}")"
          subdomain="${subdomain_value}.${fqdn}"
          subdomain_ips=$(dns_resolve "${subdomain}")
          EXIT_CODE="$?"
          if [[ -z "${subdomain_ips}" ]]
          then
            add_test_result "FAILED" "$(get_readiness_message ".connectivity.subdomain_ip_missing_error" "${MESSAGE_FILE}" "${subdomain}" )"
          else
              subdomain_resolve_map=$(echo "${subdomain_resolve_map}" |  jq --arg key "${subdomain}" --arg value "$(echo -n "${subdomain_ips}" | tr '\n' ',')" '.[$key]=($value)')
              validate_ip "${subdomain_ips}" "${subdomain}"
              if [[ $subdomain_ips != "${fqdn_ips}" ]]
              then
                add_test_result "WARN" "$(get_readiness_message ".connectivity.subdomain_error" "${MESSAGE_FILE}" "${subdomain}" "${fqdn}" )"
              else
                add_test_result "PASSED" "$(get_readiness_message ".connectivity.subdomain_info" "${MESSAGE_FILE}" "${subdomain}" )"
              fi
          fi
      done
      echo "${subdomain_resolve_map}" | jq -r 'keys[] as $k | "\($k)=\(.[$k])"' > /tmp/subdomain_resolve_map.txt
    fi
}


# Checks if the port is already in use. $1 is the port to be checked
function validate_port_connection() {
    if bash -c "exec 6<>/dev/tcp/127.0.0.1/$1" &>/dev/null; then
        add_test_result "FAILED" "$(get_readiness_message ".connectivity.port_connection_error" "${MESSAGE_FILE}" "$1" )" "$(get_readiness_mitigation ".connectivity.port_connection_error" "${MESSAGE_FILE}" "$1" )"
    else
        add_test_result "PASSED" "$(get_readiness_message ".connectivity.port_connection_info" "${MESSAGE_FILE}" "$1" )"
    fi
}

function validate_required_ports() {
    [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] && add_test_result "INFO" "$(get_readiness_message ".connectivity.port_connection_skip" "${MESSAGE_FILE}" )" && return
    [[ "$(systemctl is-enabled rke2-agent 2>/dev/null)" == "enabled" ]] && add_test_result "INFO" "$(get_readiness_message ".connectivity.port_connection_skip" "${MESSAGE_FILE}" )" && return
    validate_port_connection 80
    validate_port_connection 443
}

function validate_nameserver() {
    local exitCode=0
    # If it's getting executed as non-root user, the check does not work
    [[ "$(id -un 2>/dev/null || true)" != "root" ]] && return 0

    DNS_RESOLVER_FILE="/etc/resolv.conf"
    [[ -n "${CUSTOM_DNS_RESOLVER}" ]] && DNS_RESOLVER_FILE=${CUSTOM_DNS_RESOLVER}

    if [[ ! -s "${DNS_RESOLVER_FILE}" ]]; then
      add_test_result "FAILED" "$(get_readiness_message ".connectivity.nameserver_validation_error" "${MESSAGE_FILE}" "${DNS_RESOLVER_FILE}" )"
      return 1
    fi
    grep -E '^options' "${DNS_RESOLVER_FILE}"  | awk -F'[#;]' '{print $1}' | awk '{$1="";print $0}' | grep -q ' use-vc' && IS_DNS_TCP_ENABLED="true"
    DNS_SEARCH_DOMAINS=$(grep -E '^search' "${DNS_RESOLVER_FILE}"  | tail -n 1 | awk -F'[#;]' '{print $1}' | awk '{$1="";print $0}')

    local ip_addresses
    ip_addresses=$(< "${DNS_RESOLVER_FILE}" grep "^nameserver"| awk '{print $2}' )
    if [[ ${ip_addresses} == "" ]]; then
      add_test_result "FAILED" "$(get_readiness_message ".connectivity.nameserver_config_error" "${MESSAGE_FILE}" "${DNS_RESOLVER_FILE}" )"
      return 1
    fi

    # check if the address is only 127.0.0.0/8 CIDR or invalid ip
    local is_non_loopback_present="false"
    for ip_address in ${ip_addresses}; do
      if validate_ip "${ip_address}" "Nameserver config in ${DNS_RESOLVER_FILE}" ; then
        DNS_NAMESERVERS+=("$ip_address")
      else
        exitCode=1
      fi
      [[ ! "${ip_address}" =~ ^127.* ]] && is_non_loopback_present="true"
    done

    if [[ $is_non_loopback_present == "false" ]] ; then
      add_test_result "FAILED" "$(get_readiness_message ".connectivity.loopback_ip_found" "${MESSAGE_FILE}" "${ip_address}" "${DNS_RESOLVER_FILE}" )"
      exitCode=1
    fi
    return "$exitCode"
}

function validate_telemetry_connectivity() {
  local telemetry_opt_out
  telemetry_opt_out="$(get_optional_json_value ".telemetry_optout")"

  if [[ "${telemetry_opt_out}" == "false" ]]; then
    add_test_result "WARN" "$(get_readiness_message ".connectivity.telemetry_warn" "${MESSAGE_FILE}" )"
    return;
  fi
  
  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Prerequisites.ConnectivityCheck",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer"
            }
        }
    }
}
EOF
  }

  post_response=0
  send_telemetry_unsafe "$(data)" || post_response=$?

  if [[ ${post_response} -eq 0 ]]; then
    add_test_result "PASSED" "$(get_readiness_message ".connectivity.telemetry_info" "${MESSAGE_FILE}" "${TELEMETRY_URL}")"
  else
    add_test_result "FAILED" "$(get_readiness_message ".connectivity.telemetry_error" "${MESSAGE_FILE}" "${TELEMETRY_URL}" "${post_response}")"
  fi
}

function run_validation() {
  local tcp_ports
  local udp_ports

  tcp_ports=$(jq -r '.node_checks.tcp_ports[]' "${CONFIG}")
  udp_ports=$(jq -r '.node_checks.udp_ports[]' "${CONFIG}")
  if [[ -n $NODES ]] ; then
    #shellcheck disable=SC2005
    echo "$(get_readiness_message ".connectivity.start_info" "${MESSAGE_FILE}" "cluster connections")"
    transport_connectivity_check "tcp" "${tcp_ports}"
    transport_connectivity_check "udp" "${udp_ports}"
    echo -e "$(get_readiness_message ".connectivity.end_info" "${MESSAGE_FILE}" "cluster connections")\n"
  fi

 #shellcheck disable=SC2005
  echo "$(get_readiness_message ".connectivity.start_info" "${MESSAGE_FILE}" "nameservers")"
  if validate_nameserver; then
    add_test_result "PASSED" "Nameserver validation successful"
  fi
  echo -e "$(get_readiness_message ".connectivity.end_info" "${MESSAGE_FILE}" "nameservers")\n"

  #shellcheck disable=SC2005
  echo "$(get_readiness_message ".connectivity.start_info" "${MESSAGE_FILE}" "domains resolution")"
  validate_subdomains
  echo -e "$(get_readiness_message ".connectivity.end_info" "${MESSAGE_FILE}" "domains resolution")\n"
  #shellcheck disable=SC2005
  echo "$(get_readiness_message ".connectivity.start_info" "${MESSAGE_FILE}" "required ports")"
  validate_required_ports
  echo -e "$(get_readiness_message ".connectivity.end_info" "${MESSAGE_FILE}" "required ports")\n"

  [[ $DEPLOYMENT_TYPE == "offline" ]] && {
    add_test_result "INFO" "Skipping external connectivity checks for offline installation\n"
    return 0
  }
  #shellcheck disable=SC2005
  echo "$(get_readiness_message ".connectivity.start_info" "${MESSAGE_FILE}" "external connections")"
  outbound_connectivity_checks
  echo -e "$(get_readiness_message ".connectivity.end_info" "${MESSAGE_FILE}" "external connections")\n"  

  #shellcheck disable=SC2005
  echo "$(get_readiness_message ".connectivity.start_info" "${MESSAGE_FILE}" "telemetry connectivity")"
  validate_telemetry_connectivity
  echo -e "$(get_readiness_message ".connectivity.end_info" "${MESSAGE_FILE}" "telemetry connectivity")\n"
}


function main() {
  parse_long_args "$@"
  parse_prereq_input_file
  print_center "Connectivity Validation"
  run_validation
  print_center ""
}

main "$@"
exit "${EXIT_CODE}"