#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MESSAGE_FILE="${SCRIPT_DIR}/../readiness_message_en.json"
if [[ ! -f "${MESSAGE_FILE}" ]] ; then
  echo "Readiness message file not present"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

if [[ -f "${SCRIPT_DIR}/hardware_validation.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${SCRIPT_DIR}/hardware_validation.sh"
else
  echo "Hardware utils not found at ${SCRIPT_DIR}/hardware_validation.sh"
  exit 1
fi

EXIT_CODE=0
NODE_TYPE="none"

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Run certificates checks"
  echo
  echo "Arguments"
  echo "  -d|--debug                         Increase logging verbosity to show all debug logs"
  echo "  -h|--help                          Show this help message and exit"
  echo " -n|--node-type                       Possible values [server|agent|taskmining|gpu]. Defaults to server"
  echo  
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -n | --node-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODE_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;; 

    -*|*) # unsupported flags
      # display_usage
      # echo "Error: Unsupported flag $1"
      # exit 1
      shift
      ;;
    esac
  done
}

function validate_hardware_requirements() {
    local gpu_info
    gpu_info=$(lspci | grep -i 'NVIDIA' | tr '\n' ',')
    #shellcheck disable=SC2005
    [[ -n "${gpu_info}" ]] && echo "$(get_readiness_message ".hardware_validation.skip_info" "${MESSAGE_FILE}")" && return
    
    #shellcheck disable=SC2005
    echo "$(get_readiness_message ".hardware_validation.start_info" "${MESSAGE_FILE}")"

    cpu_validation "${NODE_TYPE}"
    memory_validation "${NODE_TYPE}"

    echo -e "$(get_readiness_message ".hardware_validation.end_info" "${MESSAGE_FILE}")\n"

    ceph_raw_device_validation "${NODE_TYPE}"
    rke2_disk_validation "${NODE_TYPE}"
    [[ "${NODE_TYPE}" == "none" || "${NODE_TYPE}" == "server" ]] && data_disk_validation

    validate_noexec_option
    validate_selinux_iscsid
}

function main() {
  parse_long_args "$@"
  parse_hardware_validation_args
  print_center "Hardware Requirements Validation"
  validate_hardware_requirements
  print_center ""
}

main "$@"
exit "${EXIT_CODE}"
