#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
SQL_SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SQL_SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MODULES_DIR=$(cd "${SQL_SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" &&  -f "${MODULES_DIR}/services_installer_utils.sh" ]] ; then
  #shellcheck disable=SC1090,SC1091
  source "${MODULES_DIR}"/services_installer_utils.sh

  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}"
  exit 1
fi

VALIDATION_SCRIPT_DIR=$(cd "${SQL_SCRIPT_DIR}/../../Services_Installer/helm/argo_app_of_apps/charts/services-preinstall-validations/scripts" && pwd)
CONTAINER_BASE_LOCALTION="/service-fabric-packaging/sql"
CONTAINER_SCRIPT_LOCATION="${CONTAINER_BASE_LOCALTION}/scripts"

if [[ ! -f "${VALIDATION_SCRIPT_DIR}/Utils.ps1" || ! -f "${VALIDATION_SCRIPT_DIR}/validation.json" || ! -f "${VALIDATION_SCRIPT_DIR}/Validate-SqlReadiness.ps1" ]] ; then
  echo "Required script to run sql validation not present at ${VALIDATION_SCRIPT_DIR}"
  exit 1
fi

SQL_CONTAINER_NAME="sql_readiness_checker"
readonly SQL_SCRIPT_DIR
readonly SQL_CONTAINER_NAME

ENABLED_SERVICES=""
DNS_SERVER=""
DEPLOYMENT_TYPE="online"
##exit code
EXIT_CODE=0

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Run sql connectivity checks between node and sql server"
  echo
  echo "Arguments"
  echo "  --install-type                     Possible values will be online or offline. Defaults to online"
  echo "  -s|--enabled-services              Comma separated list of enabled services"
  echo "  --dns-server                       DNS server ip. This will overwrite sql readiness container /etc/resolv.conf"
  echo "  -d|--debug                         Increase logging verbosity to show all debug logs"
  echo "  -h|--help                          Show this help message and exit"
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -s|--enabled-services)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local value
        value=$2
        ENABLED_SERVICES="\"$(sed -r "s/,/\",\"/g" <<< "${value}")\""
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;

    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        DEPLOYMENT_TYPE="$2"
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;

    --dns-server)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        DNS_SERVER="--dns=$2"
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;

    -*|*) # unsupported flags
      shift
      ;;
    esac
  done
}

function parse_sql_prerequisite_param() {
  if [[ "${CREATE_DB}" == "true" ]]; then
    SQL_TEST_DB_CONNECTION_STRING="$(create_sql_connection_string_from_template "dotnet" "AutomationSuite_TestDB")"
  fi

  if [[ "${DEPLOYMENT_TYPE}" == "offline" ]]; then
    DOCKER_REGISTRY_URL="localhost:${DOCKER_REGISTRY_NODEPORT}"
  fi

  SQL_PREREQ_VERSION="$(grep -i version "${SQL_SCRIPT_DIR}"/version.txt | awk -F= '{print $2}')"
  SQL_PREREQ_IMAGE_NAME="$(grep -i imagename "${SQL_SCRIPT_DIR}"/version.txt | awk -F= '{print $2}')"
  CONTAINER_IMAGE_NAME="${DOCKER_REGISTRY_URL}"/"${SQL_PREREQ_IMAGE_NAME}":"${SQL_PREREQ_VERSION}"
  
  COMMAND=()
  COMMAND+=(-CreateDB "${CREATE_DB}")
  COMMAND+=(-EnableIntegratedAuth "${ENABLESQLINTEGRATEDAUTH}")
  COMMAND+=(-TempLocation "${CONTAINER_BASE_LOCALTION}")
  COMMAND+=(-SecretLocation "/run/secrets")
  COMMAND+=(-BasePath "${CONTAINER_SCRIPT_LOCATION}")
  COMMAND+=(-EnabledServices "${ENABLED_SERVICES}")
}

function clean_files() {
  touch "${SQL_SCRIPT_DIR}"/report.json
  echo "" > "${SQL_SCRIPT_DIR}"/report.json
}

function clean_container() {
  podman secret rm -a >/dev/null
  podman rm -i "${SQL_CONTAINER_NAME}" >/dev/null
}

function toUppercase() {
  local value=$1
  local upperCaseValue
  upperCaseValue="${value^^}"
  echo "${upperCaseValue}"
}

function init_krb_file() {
local upperCaseValue krbFile
upperCaseValue=$(toUppercase "${ADDOMAIN}")
read -r -d "" krbFile <<EOF
[libdefaults]
    ticket_lifetime = 8h
    rdns = false
    default_realm = ${upperCaseValue}

[realms]
 ${upperCaseValue} = {
     default_domain = ${ADDOMAIN}
 }

[domain_realm]
 ${ADDOMAIN} = ${upperCaseValue}

EOF

KRB5_CNF=$(echo -n "${krbFile}" | base64 -w 0)
export KRB5_CNF="${KRB5_CNF}"
}

function add_secret_and_env() {
  local secret_name secret_value env_name
  secret_name=$1
  secret_value=$2
  env_name=$3

  if [[ -n "${secret_name}" && -n "${secret_value}" ]]; then
    echo "${secret_value}" | base64 -d | podman secret create "${secret_name}" - >/dev/null
    SECRETS_NAME+=(--secret "${secret_name}" )
  fi

  if [[ -n "${env_name}" ]]; then
    ENVIRONMENT_NAME+=(--env "${env_name}=/run/secrets/${secret_name}")
  fi
}

function create_prereq_sql_secrets() {
  local sql_host sql_port
  SECRETS_NAME=()
  ENVIRONMENT_NAME=()
  echo "${CREATE_DB}" | podman secret create sql_create_db - >/dev/null
  SECRETS_NAME+=(--secret sql_create_db )

  sql_host="$(get_optional_json_value ".sql.server_url")"
  sql_port="$(get_json_value_or_return_default_value ".sql.port" "1433")"
  if [[ -n "${sql_host}" && "${sql_host}" != "PLACEHOLDER" ]]; then
    if [[ "$sql_host" == *"\\"* ]]; then
      sql_host=$(echo "$sql_host" | cut -d "\\" -f 1)
    fi
    ENVIRONMENT_NAME+=(--env "SQL_SERVER_HOST=${sql_host}")
    ENVIRONMENT_NAME+=(--env "SQL_SERVER_PORT=${sql_port}")
  fi

  if [[ "${CREATE_DB}" == "true" ]]; then
    add_secret_and_env "sql_server_testdb_connection" "${SQL_TEST_DB_CONNECTION_STRING}" ""
    add_secret_and_env "sql_server_default_connection" "${SQL_DEFAULT_CONNECTION_STRING}" ""
  else
    if [[ ${INSTALLSTANDALONEAICENTER} == 0 ]]; then
      add_secret_and_env "orchestrator_sql_connection_string" "${ORCHESTRATOR_SQL_CONNECTION}" ""
      add_secret_and_env "orchestrator_ta_sql_connection_string" "${ORCHESTRATOR_TA_SQL_CONNECTION}" ""
      add_secret_and_env "orchestrator_upd_sql_connection_string" "${ORCHESTRATOR_UPD_SQL_CONNECTION}" ""
      add_secret_and_env "identity_sql_connection_string" "${IDENTITY_SQL_CONNECTION}" ""
      add_secret_and_env "insights_sql_connection_string" "${INSIGHTS_SQL_CONNECTION}" ""
      add_secret_and_env "oms_sql_connection_string" "${OMS_SQL_CONNECTION}" ""
      add_secret_and_env "testmanager_sql_connection_string" "${TESTMANAGER_SQL_CONNECTION}" ""
      add_secret_and_env "taskmining_sql_connection_string" "${TASKMINING_SQL_CONNECTION}" ""
      add_secret_and_env "automationhub_sql_connection_string" "${AUTOMATIONHUB_SQL_CONNECTION}" ""
      add_secret_and_env "automationops_sql_connection_string" "${AUTOMATIONOPS_SQL_CONNECTION}" ""
      add_secret_and_env "la_sql_connection_string" "${LA_SQL_CONNECTION}" ""
      add_secret_and_env "lrm_sql_connection_string" "${LRM_SQL_CONNECTION}" ""
      add_secret_and_env "location_sql_connection_string" "${LOCATION_SQL_CONNECTION}" ""
      add_secret_and_env "audit_sql_connection_string" "${AUDIT_SQL_CONNECTION}" ""
      add_secret_and_env "orchestrator_ta_enabled" "${ORCHESTRATOR_TESTAUTOMATION_ENABLED}" ""
      add_secret_and_env "orchestrator_upd_enabled" "${ORCHESTRATOR_UPDATESERVER_ENABLED}" ""
      add_secret_and_env "dataservice_sql_connection_string" "${DATASERVICE_SQL_CONNECTION}" ""
      add_secret_and_env "resourcecatalog_sql_connection_string" "${RESOURCECATALOG_SQL_CONNECTION}" ""
    fi
    add_secret_and_env "aicenter_sql_connection_string" "${AICENTER_SQL_CONNECTION}" ""
    add_secret_and_env "du_datamanager_sql_connection_string" "${DU_DATAMANAGER_SQL_CONNECTION}" ""
  fi

  if [[ "${ENABLESQLINTEGRATEDAUTH}" == "true" ]]; then
    init_krb_file || true
    ENVIRONMENT_NAME+=(--env AD_USERNAME="${ADUSERNAME}")
    ENVIRONMENT_NAME+=(--env KRB5CCNAME="/tmp/krb5cc_1000")

    add_secret_and_env "krb5.conf" "${KRB5_CNF}" "KRB5_CONFIG"
    add_secret_and_env "default.krb5.keytab" "${USERKEYTAB}" "DEFAULT_KRB5_KTNAME"
    ENVIRONMENT_NAME+=(--env KRB5_KTNAME="/run/secrets/default.krb5.keytab")

    if [[ ${INSTALLSTANDALONEAICENTER} == 0 ]]; then
      add_secret_and_env "discoverygroup.krb5.keytab" "${USERKEYTAB}" ""
      add_secret_and_env "insights.krb5.keytab" "${USERKEYTAB}" ""
      add_secret_and_env "orchestrator.krb5.keytab" "${USERKEYTAB}" ""
      add_secret_and_env "platform.krb5.keytab" "${USERKEYTAB}" ""
      add_secret_and_env "testmanager.krb5.keytab" "${USERKEYTAB}" ""
      add_secret_and_env "automationops.krb5.keytab" "${USERKEYTAB}" ""
      add_secret_and_env "dataservice.krb5.keytab" "${USERKEYTAB}" ""
    fi
    add_secret_and_env "aicenter.krb5.keytab" "${USERKEYTAB}" ""
    add_secret_and_env "documentunderstanding.krb5.keytab" "${USERKEYTAB}" ""
  fi
}

function consolidated_report() {
  local services results result length index message level test_mitigation
  services=$(jq -r ". | keys[]" "${SQL_SCRIPT_DIR}"/report.json)
  results=$(jq -r "." "${SQL_SCRIPT_DIR}"/report.json)
  for service in ${services}; do
    length=$(echo "${results}"| jq -c ".${service}| length")
    index=0
    while (( index < length )); do
      result=$(echo "${results}"| jq -c ".${service}[$index]")
      level=$(echo "${result}" | jq -rc '.| select(.TestName).Level')
      message=$(echo "${result}" | jq -rc '.| select(.TestName).Message')
      test_mitigation=$(echo "${result}" | jq -rc '.| select(.TestName).TestMitigation')
      add_test_result "${level}" "${message}" "${test_mitigation}"
      index=$((index + 1))
    done
  done
}

function add_additional_ca_cert() {
  ADDITIONAL_CA_CERTS="$(get_optional_json_value ".additional_ca_certs")"

  if [[ -f "${ADDITIONAL_CA_CERTS}" ]]; then
    cert=$(base64 -w 0 "${ADDITIONAL_CA_CERTS}")
    add_secret_and_env "additional_ca_cert" "${cert}" ""
  fi
}

function run_validation() {
  local container_run_status=0
  local container_exit_code=0
  local container_exit_status=0

  if [[ ${EXIT_CODE} == 0 ]]; then
    \cp -fr "${VALIDATION_SCRIPT_DIR}" "${SQL_SCRIPT_DIR}"

    # As of podman version 3.2.3 podman run does not take --tls-verify. So explicitly pulling before run
    if [[ "${DEPLOYMENT_TYPE}" == "offline" ]]; then
      podman image exists "${CONTAINER_IMAGE_NAME}" || podman pull "${CONTAINER_IMAGE_NAME}" --tls-verify=false >/dev/null
      # shellcheck disable=SC2181
      if [[ $? != 0 ]]; then
        add_test_result "WARN" "Image needed for SQL validation not present and could not be pulled from registry. Please re-run this script by passing the valid offline bundle path"
        add_test_result "FAILED" "Sql validation exit with invalid status code"
        return
      fi
    fi

    # shellcheck disable=SC2086
    podman run --name "${SQL_CONTAINER_NAME}" --replace -q --network=host --mount type=bind,src="${SQL_SCRIPT_DIR}",dst="${CONTAINER_BASE_LOCALTION}",relabel=shared "${SECRETS_NAME[@]}" "${ENVIRONMENT_NAME[@]}" ${DNS_SERVER} "${CONTAINER_IMAGE_NAME}" pwsh -c "${CONTAINER_SCRIPT_LOCATION}/Validate-SqlReadiness.ps1 ${COMMAND[*]}" >/dev/null || true
    podman container exists "${SQL_CONTAINER_NAME}"
    container_run_status=$?
    if [[ "${container_run_status}" == "0" ]]; then
        container_exit_code=$(podman inspect "${SQL_CONTAINER_NAME}" --format='{{.State.ExitCode}}')
        container_exit_status=$(podman container ls -a  --format "{{.Exited}}" --filter "name=sql_readiness_checker")
    fi
    if [[ "${container_run_status}" == "0" && "${container_exit_code}" == "0" && "${container_exit_status}" == "true" ]]; then
        consolidated_report
    else
        add_test_result "FAILED" "Sql validation exit with invalid status code"
    fi
  fi
}

function main() {
  print_center "Sql Connectivity Validation"
  clean_container
  clean_files
  echo "Validating sql connectivity... STARTED"
  parse_long_args "$@"
  set -e
  if [[ ${INSTALLSTANDALONEAICENTER} == 1 ]]; then
    parse_sf_input_file
    parse_aicenter_input_file
    parse_documentunderstanding_input_file
  else  
    parse_service_input_file
  fi
  parse_sql_prerequisite_param
  set +e
  create_prereq_sql_secrets
  add_additional_ca_cert
  run_validation
  echo -e "Validating sql connectivity... DONE\n"
  clean_container
  print_center ""
}

main "$@"
exit "${EXIT_CODE}"
