#!/bin/bash


SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)

 if [[ -z "${SCRIPT_DIR}" ]]; then
     echo "Script dir not present"
     exit 1
 fi

 #shellcheck disable=SC1090,SC1091
 source "${SCRIPT_DIR}/utils.sh"

# Gets the total no of GPU enabled nodes in the cluster
function get_total_gpu_nodes() {
    nodes=$(kubectl get node --no-headers -o custom-columns=NAME:.metadata.name)
    TOTAL_GPU_NODES=0
    for node in $nodes; do
        GPU_CAPACITY=$(kubectl describe node "$node" | grep Capacity -A 7 | grep nvidia.com/gpu | cut -d: -f2 | sed 's/^[ \t]*//;s/[ \t]*$//')
        if [[ "${GPU_CAPACITY}" != "" ]];
        then
            TOTAL_GPU_NODES=$((TOTAL_GPU_NODES+1))
            export GPU_NODE_NAME=${node}
        fi
    done
    export TOTAL_GPU_NODES
}

function create_gpu_test_job() {
    kubectl delete -f ~/scripts/gpu-test-pod.yaml
    kubectl create -f ~/scripts/gpu-test-pod.yaml
    kubectl wait --for=condition=Complete job/sf-gpu-test --timeout=180s

    if [[ "$(kubectl get job sf-gpu-test -o jsonpath='{.status.conditions[?(@.type=="Complete")].status}')" == "True" ]];
    then
        echo "GPU Test Passed"
        mark_test_passed 1619695 /tmp/primary_test_results.json
        mark_test_passed 1530717 /tmp/primary_test_results.json
    else
        echo "GPU Test Failed"
        mark_test_failed 1619695 /tmp/primary_test_results.json
        mark_test_failed 1530717 /tmp/primary_test_results.json
    fi
    kubectl delete -f ~/scripts/gpu-test-pod.yaml
}

function run_gpu_script_on_server() {
    cat <<EOF > gpu_containerd.sh
if ! nvidia-smi &>/dev/null;
then
  echo "GPU Drivers are not installed on the VM. Please refer the documentation."
  exit 0
fi
if ! which nvidia-container-runtime &>/dev/null;
then
  echo "Nvidia container runtime is not installed on the VM. Please refer the documentation."
  exit 0 
fi
grep "nvidia-container-runtime" /var/lib/rancher/rke2/agent/etc/containerd/config.toml &>/dev/null && info "GPU containerd changes already applied" && exit 0
awk '1;/plugins.cri.containerd]/{print "  default_runtime_name = \"nvidia-container-runtime\""}' /var/lib/rancher/rke2/agent/etc/containerd/config.toml > /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
echo -e '\n[plugins.linux]\n  runtime = "nvidia-container-runtime"' >> /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
echo -e '\n[plugins.cri.containerd.runtimes.nvidia-container-runtime]\n  runtime_type = "io.containerd.runc.v2"\n  [plugins.cri.containerd.runtimes.nvidia-container-runtime.options]\n    BinaryName = "nvidia-container-runtime"' >> /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
EOF
    if bash gpu_containerd.sh | grep "GPU Drivers are not installed on the VM"
    then
        mark_test_passed 1661195 /tmp/primary_test_results.json
    else
        mark_test_failed 1661195 /tmp/primary_test_results.json
    fi
}

function schedule_pod_on_gpu_node_without_toleration() {
    cat <<EOF > pod_with_nodeselector.yaml
apiVersion: v1
kind: Pod
metadata:
  name: gpu-test-pod
  labels:
    env: gpu-test-pod
    app.kubernetes.io/component: test
spec:
  containers:
  - name: gpu-test-pod
    image: gpu-test-pod
    imagePullPolicy: IfNotPresent
  nodeSelector:
    kubernetes.io/hostname: ${GPU_NODE_NAME}
EOF
    kubectl delete -f pod_with_nodeselector.yaml
    kubectl apply -f pod_with_nodeselector.yaml
    sleep 5
    if kubectl describe pod gpu-test-pod | grep "had taint {nvidia.com/gpu: present}, that the pod didn't tolerate";
    then
        mark_test_passed 1525196 /tmp/primary_test_results.json
    else
        mark_test_failed 1525196 /tmp/primary_test_results.json
    fi
    kubectl delete -f pod_with_nodeselector.yaml
}

function execute_gpu_tests() {
    create_gpu_test_job
    run_gpu_script_on_server
    schedule_pod_on_gpu_node_without_toleration
}

function main() {
    get_total_gpu_nodes
    if [[ "${TOTAL_GPU_NODES}" -gt 0 ]];
    then
        execute_gpu_tests
    fi
}

main