#!/bin/bash


if ! test -d /opt/UiPathAutomationSuite/UiPath_Installer ; then
  echo "Dir: /opt/UiPathAutomationSuite/UiPath_Installer does not exist"
fi
cd /opt/UiPathAutomationSuite/UiPath_Installer || exit 1

export KUBECONFIG=/etc/rancher/rke2/rke2.yaml PATH=$PATH:/var/lib/rancher/rke2/bin

function k8s_convert_fix_point_number() {
  local k8s_number="${1}"
  local k8s_integer
  k8s_integer=$(echo "${k8s_number}" | numfmt --from=auto)
  echo "${k8s_integer}"
}

function get_current_pvc_size() {
  local pvc_labels="$1"
  local pvc_namespace="$2"
  local pvc_size
  pvc_size=$(kubectl -n "${pvc_namespace}" get pvc -l "${pvc_labels}" -o json | jq -r '([.items[].status.capacity.storage] | unique) as $stg_array | if ($stg_array | length) == 1 then $stg_array[0] else "fail" end')
  if [[ $pvc_size == "fail" ]] ; then
    echo "PVC size not consistant for same component and sub-component"
    exit 1
  else
    echo "${pvc_size}"
  fi
}

test_results_json=[]

function add_test_result() {
  local test_id="$1"
  local test_result="$2"
  test_results_json=$(echo "${test_results_json}" | jq --arg testId "${test_id}" --argjson testResult "${test_result}" '. += [{"test" : $testId, "result": $testResult}]')
}

################ PVC resize tests ##################
COMPONENT="monitoring"
SUB_COMPONENT="alertmanager"
PVC_LABELS="app.kubernetes.io/name=alertmanager"
PVC_NAMESPACE=cattle-monitoring-system
PVC_SIZE_CURRENT=$(get_current_pvc_size "${PVC_LABELS}" "${PVC_NAMESPACE}")
## 1073741824 is 1Gi
PVC_SIZE_DELTA="1073741824"



## 1619813: Running PVC resize with same size should be idempotent
timeout 1m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${PVC_SIZE_CURRENT}" --sub-component "${SUB_COMPONENT}" > /tmp/pvc-resize-script.log 2>&1
exitCode="$?"
if [[ $exitCode -eq 0 ]] ; then
  if grep --quiet 'Resizing not required for PVC' /tmp/pvc-resize-script.log ; then
    echo "[Pass] Running PVC resize with same size is idempotent"
    add_test_result "1619813" "true"
  else
    echo "[Fail] Running PVC resize with same size is not idempotent"
    add_test_result "1619813" "false"
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running PVC resize with same size"
  add_test_result "1619813" "false"
else
  echo "[Fail] Error: Running PVC resize with same size"
  add_test_result "1619813" "false"
fi


## 1619812: Running PVC resize without -f or --force operation should take user permission before proceed
PVC_SIZE_CURRENT_BYTES=$(k8s_convert_fix_point_number "${PVC_SIZE_CURRENT}")
NEW_PVC_SIZE_BYTES=$((PVC_SIZE_CURRENT_BYTES+PVC_SIZE_DELTA))
NEW_PVC_SIZE=$(echo "${NEW_PVC_SIZE_BYTES}" | numfmt  --to=iec-i)
(timeout 1m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" --sub-component "${SUB_COMPONENT}" > /tmp/pvc-resize-script.log 2>&1) &
script_pid="$!"
(timeout 1m tail -F /tmp/pvc-resize-script.log | grep --quiet --line-buffered "If you wish to continue, type 'yes'" -m1) &
grep_pid="$!"
wait "$grep_pid"
exitCode="$?"
if [[ $exitCode -eq 0 ]] ; then
  echo "[Pass] Running PVC resize without -f or --force operation took user permission before proceed"
  add_test_result "1619812" "true"
  if test -d /proc/${script_pid} ; then
    kill -INT $script_pid
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running PVC resize without -f or --force operation take user permission before proceed"
  add_test_result "1619812" "false"
else
  echo "[Fail] Error: Running PVC resize without -f or --force operation take user permission before proceed"
  add_test_result "1619812" "false"
fi

## 1619814: Running PVC resize without subcomponent should exit with Error message
timeout 1m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" > /tmp/pvc-resize-script.log 2>&1
exitCode="$?"
if [[ $exitCode -eq 1 ]] ; then
  if grep --quiet "Error: Required argument  -b | --sub-component missing" /tmp/pvc-resize-script.log ; then
    echo "[Pass] Running PVC resize without subcomponent exited with Error message"
    add_test_result "1619814" "true"
  else
    echo "[Fail] Running PVC resize without subcomponent exited with Error message but did not throw correct message"
    add_test_result "1619814" "false"
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running PVC resize without subcomponent should exit with Error message"
  add_test_result "1619814" "false"
else
  echo "[Fail] Error: Running PVC resize without subcomponent should exit with Error message"
  add_test_result "1619814" "false"
fi



## 1619815: Running PVC resize with --help argument should show required option, debug option, examples, supported subcomponent, size, force
timeout 1m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc -h > /tmp/pvc-resize-script.log 2>&1
exitCode="$?"
if [[ $exitCode -eq 0 ]] ; then
  fail_checks=()
  ## Examples check
  example_check_status=fail
  grep --quiet 'Examples'  /tmp/pvc-resize-script.log && example_check_status=pass
  [[ $example_check_status == "fail" ]] && fail_checks+=("Example not found in help")

  ## required argument check
  req_arg_check_status=fail
  grep --quiet 'Required'  /tmp/pvc-resize-script.log && req_arg_check_status=pass
  [[ $req_arg_check_status == "fail" ]] && fail_checks+=("Required argument(s) not found in help")

  ## debug arg check
  debug_arg_check_status=fail
  grep --quiet -E 'logging.*debug'  /tmp/pvc-resize-script.log && debug_arg_check_status=pass
  [[ $debug_arg_check_status == "fail" ]] && fail_checks+=("Debug argument not found in help")

  ## supported sub-component check
  supported_sub_component_check_status=fail
  grep --quiet -E 'Required.*Possible Values \[[a-zA-Z]+(\|[a-zA-Z]+)?\]' /tmp/pvc-resize-script.log && supported_sub_component_check_status=pass
  [[ $supported_sub_component_check_status == "fail" ]] && fail_checks+=("Supported sub-component not found in help")

  ## size argument check
  size_arg_check_status=fail
  grep --quiet '\[Required\] : Size of PVC. Allowed formats are' /tmp/pvc-resize-script.log && size_arg_check_status=pass
  [[ $size_arg_check_status == "fail" ]] && fail_checks+=("Size argument not found in help")

  ## force arg check
  force_arg_check_status=fail
  grep --quiet 'force' /tmp/pvc-resize-script.log && force_arg_check_status=pass
  [[ $force_arg_check_status == "fail" ]] && fail_checks+=("Force argument not found in help")

  if [[ ${#fail_checks[@]} != 0 ]] ; then
    echo "[Fail] Running PVC resize with --help argument does not show below required information"
    printf '%s\n' "${fail_checks[@]}"
    add_test_result "1619815" "false"
  else
    echo "[Pass] Running PVC resize with --help argument shows required option, debug option, examples, supported subcomponent, size, force"
    add_test_result "1619815" "true"
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running PVC resize with --help argument should show required option, debug option, examples, supported subcomponent, size, force"
  add_test_result "1619815" "false"
else
  echo "[Fail] Error: Running PVC resize with --help argument should show required option, debug option, examples, supported subcomponent, size, force"
  add_test_result "1619815" "false"
fi 

## 1619816: Running PVC resize with size greater that PVC node storage capacity should exit
## For simplicity , we will use overprovisoning percentage directly with mount point /datadisk (without taking into account storageReserved)
LONGHORN_NAMESPACE=longhorn-system
LONGHORN_MANGER_LABELS="app=longhorn-manager"
# shellcheck disable=SC2046
over_provisioning_percentage=$(kubectl -n ${LONGHORN_NAMESPACE}  exec $(kubectl -n ${LONGHORN_NAMESPACE}  get po  -l "${LONGHORN_MANGER_LABELS}"  --field-selector status.phase=Running -o jsonpath='{.items[0].metadata.name}') -- curl -s  "http://longhorn-backend.${LONGHORN_NAMESPACE}:9500/v1/settings/storage-over-provisioning-percentage" | jq -r '.value')
datadisk_size=$(findmnt -n -o size --target /datadisk)
datadisk_size_bytes=$(k8s_convert_fix_point_number "${datadisk_size}")
hard_limit_bytes=$((datadisk_size_bytes * over_provisioning_percentage / 100))
hard_limit=$(echo "${hard_limit_bytes}" | numfmt  --to=iec-i)

timeout 1m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${hard_limit}" --sub-component "${SUB_COMPONENT}" > /tmp/pvc-resize-script.log 2>&1
exitCode="$?"
if [[ $exitCode -eq 1 ]] ; then 
  if grep --quiet 'Error: Aborting expansion, Node(s) do not have enough storage capacity to expand the requested PVC(s)' /tmp/pvc-resize-script.log ; then
    echo "[Pass] Running PVC resize with size greater that PVC node storage capacity exits"
    add_test_result "1619816" "true"
  else
    echo "[Fail] Running PVC resize with size greater that PVC node storage capacity does not exit"
    add_test_result "1619816" "false"
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running PVC resize with size greater that PVC node storage capacity should exit"
  add_test_result "1619816" "false"
else
  echo "[Fail] Error: Running PVC resize with size greater that PVC node storage capacity should exit"
  add_test_result "1619816" "false"
fi

## 1619820 Running PVC resize command with wrong argument name should show proper message
timeout 1m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${hard_limit}" --sub-component "${SUB_COMPONENT}" --wrong-arg > /tmp/pvc-resize-script.log 2>&1
exitCode="$?"
if [[ $exitCode -eq 1 ]] ; then
  if grep --quiet 'Error: Unsupported flag --wrong-arg' /tmp/pvc-resize-script.log ; then
    echo "[Pass] Running PVC resize command with wrong argument name shows proper message"
    add_test_result "1619820" "true"
  else
    echo "[Fail] Running PVC resize command with wrong argument name does not show proper message"
    add_test_result "1619820" "false"
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running PVC resize command with wrong argument name should show proper message"
  add_test_result "1619820" "false"
else
  echo "[Fail] Error: Running PVC resize command with wrong argument name should show proper message"
  add_test_result "1619820" "false"
fi          

## 1619821 Run pvc resize for a pod which is created in agent node
## 1619811 Running PVC resize with -f or --force option should show force run info
## 1619819 Running PVC resize should increase/decrease size of PVC
## 1619818 Running PVC resize for for sub-component should not changes size of other component/sub-component
kubectl get pvc -A -o json  | jq '[.items[] | select(.metadata.namespace == "cattle-monitoring-system" and  .metadata.labels["app.kubernetes.io/name"] == "alertmanager" | not) | {name: .metadata.name, namespace: .metadata.namespace, size: .status.capacity.storage}] | sort_by(.name) | .[]' > /tmp/pre-resize-state.json
ready_agent_node=$(kubectl get node -o json   | jq -r '[.items[] | select((.status.conditions[] | select(.type=="Ready" and .status=="True")) and (.metadata.labels["node-role.kubernetes.io/master"] == null or .metadata.labels["node-role.kubernetes.io/master"] == false)) | .metadata.labels["kubernetes.io/hostname"]] as $ready_agent_node | if ($ready_agent_node | length) > 0 then $ready_agent_node[0] else "no-ready-agent-node-found"end')
if [[ $ready_agent_node == "no-ready-agent-node-found" ]] ; then
  echo "[Skipped] Run pvc resize for a pod which is created in agent node"
  timeout 10m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" --sub-component "${SUB_COMPONENT}" --force > /tmp/pvc-resize-script.log 2>&1
  exitCode="$?"
  if [[ $exitCode -eq 0 ]] ; then
    fail_checks=()
    if grep --quiet 'ROLLING BACK' /tmp/pvc-resize-script.log ; then
      fail_checks+=("[Fail] Running PVC resize failed due to timeout.'ROLLING BACK' found")
    fi

    if ! grep --quiet 'Successfully resized PVC' /tmp/pvc-resize-script.log ; then
      fail_checks+=("[Fail] Running PVC resize does not increase size")
    fi

    if ! grep --quiet 'Successfully attached PVC' /tmp/pvc-resize-script.log ; then
      fail_checks+=("[Fail] Running PVC resize does not attach back the pvc")
    fi

    if [[ ${#fail_checks[@]} != 0 ]] ; then
      printf '%s\n' "${fail_checks[@]}"
      add_test_result "1619819" "false"
    else
      echo "[Pass] Running PVC resize increases size of PVC"
      add_test_result "1619819" "true"
    fi
  elif [[ $exitCode -eq 124 ]] ; then
    echo "[Fail] Timeout: Running PVC resize should increase/decrease size of PVC"
    add_test_result "1619819" "false"
  else
    echo "[Fail] Error: Running PVC resize should increase/decrease size of PVC"
    add_test_result "1619819" "false"
  fi
else
  istio_proxy_image=$(kubectl -n istio-system get ds istio-ingressgateway -o json | jq -r '.spec.template.spec.containers[] | select(.name=="istio-proxy").image')
  # shellcheck disable=SC2086
  echo '{"apiVersion": "apps/v1", "kind": "StatefulSet", "metadata": {"name": "dummy-pvc", "namespace": "'${PVC_NAMESPACE}'"}, 
  "spec": {"replicas": 1, "selector": {"matchLabels": {"app": "dummy-pvc", "app.kubernetes.io/name": "alertmanager"} }, 
  "serviceName": "dummy-pvc", "template": {"metadata": {"labels": {"app": "dummy-pvc", "app.kubernetes.io/name": "alertmanager"} },
  "spec": {"containers": [{"command": ["/bin/sh", "-c", "tail -f /dev/null"], "image": "'${istio_proxy_image}'", "imagePullPolicy": "IfNotPresent",
  "name": "dummy-pvc", "volumeMounts": [{"mountPath": "/usr/share/nginx/html", "name": "dummy"} ] } ], 
  "nodeSelector": {"kubernetes.io/hostname": "'${ready_agent_node}'"}, "terminationGracePeriodSeconds": 10, 
  "tolerations": [{"operator": "Exists"} ] } }, "volumeClaimTemplates": [{"metadata": {"name": "dummy"}, 
  "spec": {"accessModes": ["ReadWriteOnce"], "resources": {"requests": {"storage": "1Gi"} }, 
  "storageClassName": "longhorn-backup-single-replica"} } ] } }' | jq '.' > /tmp/dummy-agent-pvc-pod.json
  kubectl apply -f /tmp/dummy-agent-pvc-pod.json
  if kubectl -n "${PVC_NAMESPACE}"  wait pod -l app=dummy-pvc  --for=condition=Ready --timeout=2m ; then
    timeout 10m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" --sub-component "${SUB_COMPONENT}" --force > /tmp/pvc-resize-script.log 2>&1
    exitCode="$?"
    kubectl delete -f /tmp/dummy-agent-pvc-pod.yaml
    kubectl -n "${PVC_NAMESPACE}" wait pod -l app=dummy-pvc --for=delete --timeout=2m
    kubectl -n "${PVC_NAMESPACE}" delete pvc dummy-dummy-pvc-0
    if [[ $exitCode -eq 0 ]] ; then
      fail_checks=()
      if grep --quiet 'ROLLING BACK' /tmp/pvc-resize-script.log ; then
        fail_checks+=("[Fail] Running PVC resize failed due to timeout.'ROLLING BACK' found")
      fi

      if ! grep --quiet 'Successfully resized PVC' /tmp/pvc-resize-script.log ; then
        fail_checks+=("[Fail] Running PVC resize does not increase size")
      fi

      if ! grep --quiet 'Successfully attached PVC' /tmp/pvc-resize-script.log ; then
        fail_checks+=("[Fail] Running PVC resize does not attach back the pvc")
      fi

      if [[ ${#fail_checks[@]} != 0 ]] ; then
        printf '%s\n' "${fail_checks[@]}"
        add_test_result "1619819" "false"
        add_test_result "1619821" "false"
      else
        echo "[Pass] Running pvc resize for a pod which is created on agent node executed successfully"
        echo "[Pass] Running PVC resize increases size of PVC"
        add_test_result "1619819" "true"
        add_test_result "1619821" "true"
      fi
    elif [[ $exitCode -eq 124 ]] ; then
      echo "[Fail] Timeout: Run pvc resize for a pod which is created in agent node"
      echo "[Fail] Timeout: Running PVC resize should increase/decrease size of PVC"
      add_test_result "1619819" "false"
      add_test_result "1619821" "false"
    else
      echo "[Fail] Error: Run pvc resize for a pod which is created in agent node"
      echo "[Fail] Error: Running PVC resize should increase/decrease size of PVC"
      add_test_result "1619819" "false"
      add_test_result "1619821" "false"
    fi
  else
    echo "[Fatal] Timeout: waiting for pod on agent node to go into ready state"
    echo "Please check why pod did not come-up in stipulated time"
  fi
fi
## Force check
if grep --quiet 'Proceeding to resize PVC(s) with force' /tmp/pvc-resize-script.log ; then
  echo "[Pass] Running PVC resize with -f or --force shows force run info"
  add_test_result "1619811" "true"
else
  echo "[Fail] Running PVC resize with -f or --force option does not show force run info"
  add_test_result "1619811" "false"
fi
kubectl get pvc -A -o json  | jq '[.items[] | select(.metadata.namespace == "cattle-monitoring-system" and  .metadata.labels["app.kubernetes.io/name"] == "alertmanager" | not) | {name: .metadata.name, namespace: .metadata.namespace, size: .status.capacity.storage}] | sort_by(.name) | .[]' > /tmp/post-resize-state.json
if diff  -q /tmp/pre-resize-state.json /tmp/post-resize-state.json  >> /dev/null 2>&1 ; then
  echo "[Pass] Running PVC resize for sub-component does not changes size of other component/sub-component"
  add_test_result "1619818" "true"
else
  echo "[Fail] Running PVC resize for sub-component changes size of other component/sub-component"
  add_test_result "1619818" "false"
fi


## 1619822: Run resize for object-store pvc
COMPONENT="objectstore"
SUB_COMPONENT="data"
PVC_LABELS="ceph.rook.io/DeviceSet"
PVC_NAMESPACE=rook-ceph
PVC_SIZE_CURRENT=$(get_current_pvc_size "${PVC_LABELS}" "${PVC_NAMESPACE}")
PVC_SIZE_CURRENT_BYTES=$(k8s_convert_fix_point_number "${PVC_SIZE_CURRENT}")
NEW_PVC_SIZE_BYTES=$((PVC_SIZE_CURRENT_BYTES+PVC_SIZE_DELTA))
NEW_PVC_SIZE=$(echo "${NEW_PVC_SIZE_BYTES}" | numfmt  --to=iec-i)

## know issue https://docs.uipath.com/automation-suite/docs/troubleshooting#failure-to-resize-objectstore-pvc
ROOK_CEPH_OSD_PREPARE=$(kubectl -n "${PVC_NAMESPACE}" get pods | grep rook-ceph-osd-prepare-set | awk '{print $1}')
if [[ -n ${ROOK_CEPH_OSD_PREPARE} ]]; then
  for pod in ${ROOK_CEPH_OSD_PREPARE}; do
    kubectl -n "${PVC_NAMESPACE}" delete pod "$pod"
  done;
fi

timeout 10m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" --sub-component "${SUB_COMPONENT}" --force > /tmp/pvc-resize-script.log 2>&1
exitCode="$?"
if [[ $exitCode -eq 0 ]] ; then
  fail_checks=()
  if grep --quiet 'ROLLING BACK' /tmp/pvc-resize-script.log ; then
    fail_checks+=("[Fail] Running resize for object-store pvc due to 'ROLLING BACK'")
  fi
          
  if ! grep --quiet 'Successfully resized PVC' /tmp/pvc-resize-script.log  ; then
    fail_checks+=("[Fail] Running resize for object-store pvc as 'Successfully resized PVC' pattern not found")
  fi

  if ! grep --quiet 'Successfully attached PVC' /tmp/pvc-resize-script.log  ; then
    fail_checks+=("[Fail] Running resize for object-store pvc as 'Successfully attached PVC' pattern not found")
  fi

  if [[ ${#fail_checks[@]} != 0 ]] ; then
    printf '%s\n' "${fail_checks[@]}"
    add_test_result "1619822" "false"
  else
    echo "[Pass] Running resize for object-store pvc"
    add_test_result "1619822" "true"
  fi
elif [[ $exitCode -eq 124 ]] ; then
  echo "[Fail] Timeout: Running resize for object-store pvc"
  add_test_result "1619822" "false"
else
  echo "[Fail] Error: Running resize for object-store pvc"
  add_test_result "1619822" "false"
fi


## 1619817 Connection/Session timeout while running PVC resize should be auto recoverable
COMPONENT="monitoring"
SUB_COMPONENT="alertmanager"
PVC_LABELS="app.kubernetes.io/name=alertmanager"
PVC_NAMESPACE=cattle-monitoring-system
PVC_SIZE_DELTA="1073741824"
PVC_SIZE_CURRENT=$(get_current_pvc_size "${PVC_LABELS}" "${PVC_NAMESPACE}")
PVC_SIZE_CURRENT_BYTES=$(k8s_convert_fix_point_number "${PVC_SIZE_CURRENT}")
NEW_PVC_SIZE_BYTES=$((PVC_SIZE_CURRENT_BYTES+PVC_SIZE_DELTA))
NEW_PVC_SIZE=$(echo "${NEW_PVC_SIZE_BYTES}" | numfmt  --to=iec-i)        

(timeout 10m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" --sub-component "${SUB_COMPONENT}" --force > /tmp/pvc-resize-script.log 2>&1) &
script_pid="$!"
## Sleep random seconds upto 60
sleep $(( ( RANDOM % 60 )  + 1 ))
if test -d /proc/${script_pid} ; then
  kill -INT  $script_pid
  timeout 10m ./configureUiPathAS.sh "${COMPONENT}" resize-pvc --size "${NEW_PVC_SIZE}" --sub-component "${SUB_COMPONENT}" --force > /tmp/pvc-resize-script.log 2>&1
  exitCode="$?"
  if [[ $exitCode -eq 0 ]] ; then
    echo "[Pass] Connection/Session timeout while running PVC resize recoverabled automatically"
    add_test_result "1619817" "true"
  elif [[ $exitCode -eq 124 ]] ; then
    echo "[Fail] Timeout: Connection/Session timeout while running PVC resize should be auto recoverable"
    add_test_result "1619817" "false"
  else
    echo "[Fail] Error: Connection/Session timeout while running PVC resize should be auto recoverable"
    add_test_result "1619817" "false"
  fi
else
  wait $script_pid
  echo "[Skipped] Connection/Session timeout while running PVC resize should be auto recoverable. Process exited with status code $? before simulting interrupt"
fi

echo "${test_results_json}" | jq '.' > /tmp/pvc_test_results.json