#!/bin/bash

function export_s3_config() {
    OBJECT_GATEWAY_INTERNAL_IP="rook-ceph-rgw-rook-ceph"
    OBJECT_GATEWAY_INTERNAL_HOST=$(kubectl -n rook-ceph get services/$OBJECT_GATEWAY_INTERNAL_IP -o jsonpath="{.spec.clusterIP}")
    OBJECT_GATEWAY_INTERNAL_PORT=$(kubectl -n rook-ceph get services/$OBJECT_GATEWAY_INTERNAL_IP -o jsonpath="{.spec.ports[0].port}")
    STORAGE_ACCESS_KEY=$(kubectl -n rook-ceph get secret ceph-object-store-secret -o json | jq '.data.OBJECT_STORAGE_ACCESSKEY' | sed -e 's/^"//' -e 's/"$//' | base64 -d)
    STORAGE_SECRET_KEY=$(kubectl -n rook-ceph get secret ceph-object-store-secret -o json | jq '.data.OBJECT_STORAGE_SECRETKEY' | sed -e 's/^"//' -e 's/"$//' | base64 -d)
    export AWS_HOST=$OBJECT_GATEWAY_INTERNAL_HOST
    export AWS_ENDPOINT=$OBJECT_GATEWAY_INTERNAL_HOST:$OBJECT_GATEWAY_INTERNAL_PORT
    export AWS_ACCESS_KEY_ID=$STORAGE_ACCESS_KEY
    export AWS_SECRET_ACCESS_KEY=$STORAGE_SECRET_KEY
}

function trigger_ceph_functional_tests() {
    export_s3_config
    filename=$(date '+%d%m%Y%H%M%S')
    dd if=/dev/zero of=/mnt/"$filename".txt count=10 bs=1048576
    bucket_name="s3://ha-test-bucket"
    
    if ! /usr/local/bin/s3cmd mb --host="${AWS_HOST}" --host-bucket=  "$bucket_name" --no-ssl; then 
        return 1
    elif ! /usr/local/bin/s3cmd put /mnt/"$filename".txt --no-ssl --host="${AWS_HOST}" --host-bucket=  "$bucket_name"/"$filename".txt; then
        return 1
    elif ! /usr/local/bin/s3cmd get "$bucket_name"/"$filename".txt --recursive --no-ssl --force --host="${AWS_HOST}" --host-bucket=  "$bucket_name"; then
        return 1
    elif ! /usr/local/bin/s3cmd ls "$bucket_name"/"$filename".txt --host="${AWS_HOST}" --host-bucket=  "$bucket_name" --no-ssl; then
        return 1
    else
        return 0
    fi
}

function testRookCephPodCount() {
    local expectedOSDMONPodCount expectedMGRRGWPodCount replicaCount podCount validTestRailIds requestedTestRailIds result
    
    expectedOSDMONPodCount=$1
    expectedMGRRGWPodCount=$2
    requestedTestRailIds=$3
    validTestRailIds=("1621326" "1621327" "1663790" "1663803")

    for validTestRailId in "${validTestRailIds[@]}"; do
        if [[ "${requestedTestRailIds}" =~ "${validTestRailId}" ]]; then

            replicaCount=$(fetchReplicaCount cephcluster rook-ceph rook-ceph ".spec.mon.count")

            #case when API server is node responding i.e. 2 master nodes down
            if [[ -z $replicaCount && -z $expectedOSDMONPodCount && -z $expectedMGRRGWPodCount ]]; then
                appendOutput "${validTestRailId}" "true"
                return
            fi

            podCount=$(kubectl -n rook-ceph get pod --selector=app=rook-ceph-mon --field-selector=status.phase=Running --output json | jq -j ".items | length")
            if ! [[ ${podCount} -ge ${expectedOSDMONPodCount} && ${podCount} -le ${replicaCount} ]]; then
                appendOutput "${validTestRailId}" "false"
                return
            fi

            replicaCount=$(fetchReplicaCount cephcluster rook-ceph rook-ceph ".spec.storage.storageClassDeviceSets[0].count")
            podCount=$(kubectl -n rook-ceph get pod --selector=app=rook-ceph-osd --field-selector=status.phase=Running --output json | jq -j ".items | length")
            if ! [[ ${podCount} -ge ${expectedOSDMONPodCount} && ${podCount} -le ${replicaCount} ]]; then
                appendOutput "${validTestRailId}" "false"
                return
            fi

            replicaCount=$(kubectl -n rook-ceph get deploy --selector=app=rook-ceph-mgr -o name | wc -l)
            podCount=$(kubectl -n rook-ceph get pod --selector=app=rook-ceph-mgr --field-selector=status.phase=Running --output json | jq -j ".items | length")
            if ! [[ ${podCount} -ge ${expectedMGRRGWPodCount} && ${podCount} -le ${replicaCount} ]]; then
                appendOutput "${validTestRailId}" "false"
                return
            fi

            podCount=$(kubectl -n rook-ceph get pod --selector=app=rook-ceph-rgw --field-selector=status.phase=Running --output json | jq -j ".items | length")
            if ! [[ ${podCount} -ge ${expectedMGRRGWPodCount} ]]; then
                appendOutput "${validTestRailId}" "false"
                return
            fi

            if ! trigger_ceph_functional_tests; then 
                appendOutput "${validTestRailId}" "false"
                return
            fi

            appendOutput "${validTestRailId}" "true"
        fi
    done
}

function setRookCephValidationParam() {
    local validationStage
    validationStage=$1
    case "$validationStage" in
    3nodeup)
        monosdPodCount=3
        mgrRGWPodCount=2
        testRailIds="1621326"
        ;;
    rebootAllServers)
        monosdPodCount=3
        mgrRGWPodCount=2
        testRailIds="1663790"
        ;;
    2nodeup)
        testRailIds="1621327"
        monosdPodCount=2
        mgrRGWPodCount=2
        ;;
    1nodeup|0nodeup)
        monosdPodCount=""
        mgrRGWPodCount=""
        testRailIds="1663803"
        ;;
    *)
        echo "Unexpected option ${validationStage} passed to ${FUNCNAME[0]}"
        exit 1
        ;;
    esac
}

function executeRookCephTest() {
    setRookCephValidationParam "${VALIDATE_STAGE}"
    testRookCephPodCount "${monosdPodCount}" "${mgrRGWPodCount}" "${testRailIds}"
}