#!/bin/bash

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)

if [[ -z "${SCRIPT_DIR}" ]]; then
    echo "Script dir not present"
    exit 1
fi

ROOT_DIR="/opt/UiPathAutomationSuite/UiPath_Installer"

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}/utils.sh"

declare -A serviceNameMapping

function prepareEnvironment() {
  local installerRootDir="$1"
  local standAloneAicenter="$2"

  if [[ ! -f "${installerRootDir}/Modules/utils.sh" ]]; then
    echo "Installer utils file not present at ${installerRootDir}/Modules/utils.sh"
    exit 1
  fi

  source "${installerRootDir}/Modules/utils.sh"
  export ALL_SERVICES=("platform" "orchestrator" "automation_hub" "automation_ops" "insights" "documentunderstanding" "task_mining")

  DEFAULT_SERVICES="platform,orchestrator"
  serviceNameMapping["platform"]="platform"
  serviceNameMapping["orchestrator"]="orchestrator"
  serviceNameMapping["automation_hub"]="automationhub"
  serviceNameMapping["automation_ops"]="automationops"
  serviceNameMapping["insights"]="insights"
  serviceNameMapping["documentunderstanding"]="documentunderstanding"
  serviceNameMapping["task_mining"]="taskmining"
  serviceNameMapping["test_manager"]="testmanager"

  if [[ "${standAloneAicenter}" == "1" ]] ; then
    export ALL_SERVICES=("documentunderstanding")
    serviceNameMapping["documentunderstanding"]="documentunderstanding"
    DEFAULT_SERVICES="documentunderstanding"
  fi
}


function checkMessageIsAvailable() {
  local logFile="$1"
  local message="$2"
  local messageFindResult result
  result=1
  messageFindResult="$(grep "${message}" ${logFile})"
  if [[ -n "${messageFindResult}" ]]; then
    result=0
  fi
  echo ${result}
}

function isServiceEnabled() {
  local service="$1"
  local clusterConfigFile="$2"
  local isEnabled result
  result=0
  isEnabled="$(jq ".${service}.enabled" ${clusterConfigFile} | xargs)"
  if [[ -z "${isEnabled}" || "${isEnabled}" == "null" ]]; then
    if [[ "${DEFAULT_SERVICES}" == *"${service}"* ]]; then
      isEnabled="true"
    fi
  fi
  
  if [[ "${isEnabled}" == "true" ]]; then
    result=1
  fi
  echo ${result}
}

function checkIfMandatoryMessageAvailable() {
  local logFile="$1"
  local serviceName="$2"
  local dbConnectionStatus permissionMessageStatus result
  result=1
  dbConnectionStatus="$(grep "Able to connect with ${serviceName} database using authentication mechanism provided" ${logFile})"
  permissionMessageStatus="$(grep "SQL User have permissions to create schema (db_ddladmin), read (db_datareader) and write (db_datawriter) to database" ${logFile})"
  
  if [[ -n "${dbConnectionStatus}" && -n "${permissionMessageStatus}" ]]; then
    result=0
  fi
  echo ${result}
}

function whenSqlCredentialsIsValidAndSQLServerRechableAndCreateDBTrue() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local prerequisitesArgs=("${@:3}")
  local newClusterConfigFile result sql_url

  is_du_enabled="$(jq ".documentunderstanding.enabled" ${clusterConfigFile} | xargs)"
  is_insight_enabled="$(jq ".insights.enabled" ${clusterConfigFile} | xargs)"
  sql_url="$(jq ".sql.server_url" ${clusterConfigFile} | xargs)"

  newClusterConfigFile="${installerRootDir}"/sql_input.json
  jq ".sql += {\"create_db\": \"true\"}" ${clusterConfigFile} > ${newClusterConfigFile}

  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${newClusterConfigFile} ${prerequisitesArgs[@]} &>/tmp/sql.log
  result=$(checkIfMandatoryMessageAvailable "/tmp/sql.log" "master")

  if [[ ${result} == 0 ]]; then
    result=$(checkMessageIsAvailable "/tmp/sql.log" "SQL User have permissions to create database")
  fi
  if [[ ${result} == 0 && -n "${is_insight_enabled}" && "${is_insight_enabled}" == "true" ]]; then
    result=$(checkMessageIsAvailable "/tmp/sql.log" "SQL server compatibility level is >= 130 and Columnstore indexer, JSON support is enabled for Insights")
  fi
  if [[ ${result} == 0 && -n "${is_du_enabled}" && "${is_du_enabled}" == "true" ]]; then
    result=$(checkMessageIsAvailable "/tmp/sql.log" "Full Text Search is enabled in sql sever.")
  fi

  if [[ ${result} == 0 && -n "${sql_url}" && "${sql_url}" != "null" && "${sql_url}" != "PLACEHOLDER" ]]; then
    result=$(checkMessageIsAvailable "/tmp/sql.log" "Able to reach sql sever on given fqdn and port.")
  fi

  if [[ ${result} == 0 ]]; then
    appendOutput "1648588" "true" 
  else
    appendOutput "1648588" "false" 
  fi
}

# this test will execute only if create database is set to false in input.json
function whenSqlCredentialsIsValidAndUserIsDBOwnerAndSQLServerRechableAndCreateDBFalse() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local prerequisitesArgs=("${@:3}")
  local createDb result is_enabled

  createDb=$(jq ".sql.create_db" ${clusterConfigFile} | xargs)
  if [[ -n "${createDb}" && "${createDb}" == "true" ]]; then
    return 0
  fi

  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${clusterConfigFile} ${prerequisitesArgs[@]} &>/tmp/sql.log
  
  for service in ${ALL_SERVICES[@]}; do
    is_enabled="$(isServiceEnabled ${serviceNameMapping[${service}]} ${clusterConfigFile})"
    if [[ "${is_enabled}" == "1" ]] ; then
      result=$(checkIfMandatoryMessageAvailable "/tmp/sql.log" "${serviceNameMapping[${service}]}")
      if [[ ${result} == 0 && "${service}" == "insights" ]]; then
        result=$(checkMessageIsAvailable "/tmp/sql.log" "SQL server compatibility level is >= 130 and Columnstore indexer, JSON support is enabled for Insights")
      fi
      if [[ ${result} == 0 && "${service}" == "documentunderstanding" ]]; then
        result=$(checkMessageIsAvailable "/tmp/sql.log" "Full Text Search is enabled in sql sever.")
      fi
      if [[ ${result} == 1 ]]; then
        break;
      fi
    fi
  done

  if [[ ${result} == 0 ]]; then
    appendOutput "1660857" "true" 
  else
    appendOutput "1660857" "false" 
  fi
}

function whenSqlIsNotAccessableFromServer() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local prerequisitesArgs=("${@:3}")
  local result newClusterConfigFile
  newClusterConfigFile="${installerRootDir}"/sql_input.json

  jq ".sql += {\"server_url\": \"randome-sql.com\"}" ${clusterConfigFile} > ${newClusterConfigFile}

  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${newClusterConfigFile} ${prerequisitesArgs[@]} &>/tmp/sql.log
  
  result=$(checkMessageIsAvailable "/tmp/sql.log" "Unable to reach sql sever on given fqdn and port")
  if [[ ${result} == 0 ]]; then
    appendOutput "1660859" "true" 
  else
    appendOutput "1660859" "false" 
  fi
}

function whenSqlUserCredentialsIsInValid() {
  local installerRootDir="$1"
  local clusterConfigFile="$2"
  local userName="User Id=$3;Password"
  local prerequisitesArgs=("${@:4}")
  local sql_dotnet_template new_sql_dotnet_template createDb is_enabled newClusterConfigFile
  sql_dotnet_template=$(jq '.sql_connection_string_template' ${clusterConfigFile})
  new_sql_dotnet_template=$(echo ${sql_dotnet_template//User Id=[a-z A-Z][a-z A-Z]*;Password/$userName})
  newClusterConfigFile="${installerRootDir}/updated_input.json"
  jq ".sql_connection_string_template = ${new_sql_dotnet_template}" ${clusterConfigFile} > ${newClusterConfigFile}
  "${installerRootDir}/validateUiPathASReadiness.sh" -i ${newClusterConfigFile} ${prerequisitesArgs[@]} &>/tmp/sql.log
  createDb=$(jq ".sql.create_db" ${clusterConfigFile} | xargs)

  if [[ -n "${createDb}" && "${createDb}" == "true" ]]; then
    result=$(checkMessageIsAvailable "/tmp/sql.log" "Unable to connect with master database using authentication mechanism provided")
  else
    for service in ${ALL_SERVICES[@]}; do
      is_enabled="$(isServiceEnabled ${serviceNameMapping[${service}]} ${clusterConfigFile})"
      if [[ "${is_enabled}" == "1" && "${service}" != "documentunderstanding" ]] ; then
        result=$(checkMessageIsAvailable "/tmp/sql.log" "Unable to connect with ${serviceNameMapping[${service}]} database using authentication mechanism provided")
        if [[ ${result} == 1 ]]; then
          break;
        fi
      fi
    done
  fi

  if [[ ${result} == 0 ]]; then
    appendOutput "1660864" "true" 
  else
    appendOutput "1660864" "false" 
  fi
}

function executeTest() {
    local prerequisites_args=()
    prerequisites_args+=(--install-type "${DEPLOYMENT_TYPE}")
    if [[ -n ${NODE_TYPE} ]]; then
      prerequisites_args+=(-n "${NODE_TYPE}")
    else
      prerequisites_args+=(-n "server")
    fi

    if [[ "${DEPLOYMENT_TYPE}" == "offline" ]]; then
      prerequisites_args+=(--offline-bundle "${OFFLINE_BUNDLE_PATH}")
      prerequisites_args+=(--offline-tmp-folder "${OFFLINE_TMP_FOLDER}")
    fi

    prepareEnvironment "${ROOT_DIR}"

    whenSqlCredentialsIsValidAndSQLServerRechableAndCreateDBTrue "${ROOT_DIR}" "${ROOT_DIR}/input.json" "${prerequisites_args[@]}"
    whenSqlCredentialsIsValidAndUserIsDBOwnerAndSQLServerRechableAndCreateDBFalse  "${ROOT_DIR}" "${ROOT_DIR}/input.json" "${prerequisites_args[@]}"
    whenSqlIsNotAccessableFromServer "${ROOT_DIR}" "${ROOT_DIR}/input.json" "${prerequisites_args[@]}"
    whenSqlUserCredentialsIsInValid "${ROOT_DIR}" "${ROOT_DIR}/input.json" "random_user" "${prerequisites_args[@]}"
}

parse_args "$@"
executeTest

# fetch deployment type