#!/bin/bash

function parse_args() {
  INFILE=""
  OUTFILE=""
  VALIDATE_STAGE=""
  DEPLOYMENT_TYPE="online"
  NODE_TYPE=""
  
  while getopts 'o:s:d:n:' flag "$@"; do
    case "${flag}" in
    o)
      echo "Using output parameters file: ${OPTARG}"
      OUTFILE=${OPTARG}
      ;;
    s)
      echo "Using validation param: ${OPTARG}"
      VALIDATE_STAGE=${OPTARG}
      ;;
    d)
      echo "Using validation param: ${OPTARG}"
      DEPLOYMENT_TYPE=${OPTARG}
      ;;
    n)
      echo "Using validation param: ${OPTARG}"
      NODE_TYPE=${OPTARG}
      ;;
    *)
      echo "Unexpected option ${flag}"
      ;;
    esac
  done

  shift $((OPTIND - 1))
}

function fetchReplicaCount() {
    local replicaCount namespace name kind key
    kind=$1
    namespace=$2
    name=$3
    key="{.spec.replicas}"
    if [[ ! -z ${4} && "${4}" != " " ]]; then
        key="{$4}"
    fi

    replicaCount=$(kubectl --request-timeout=15 -n $namespace get $kind $name -o jsonpath="${key}" )
    echo $replicaCount
}

function appendOutput() {
    local id result
    id=$1
    result=$2
    jq ". += [{\"test\" : \"${id}\", \"result\": ${result}}]" ${OUTFILE} > "${SCRIPT_DIR}/tempout.json"
    echo "appendOutput-1-${id}-${result}-${OUTFILE}-${SCRIPT_DIR}" > ~/output.json
    mv "${SCRIPT_DIR}/tempout.json" "${OUTFILE}"
}

function error_without_exit() {
  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
}

function failure_output() {
  local title=$1
  shift
  local failures=("$@")
  printf '=%.0s' {1..100}
  echo
  error_without_exit "$title"
  for error in "${failures[@]}"
  do
    [[ -n ${error} ]] && error_without_exit ">>> $error"
  done
  printf '=%.0s' {1..100}
  echo
  exit 1
}

function is_deployment_contain_keyword() {
  local deployment_name=$1
  local keyword=$2

  grep_result="$(kubectl get deploy ${deployment_name} -n uipath -o json | grep "${keyword}")"
  if [[ -n "${grep_result}" ]]; then 
    echo "true"
  else
    echo "false"
  fi
}

function array_contains_element() {
  declare -n array=$1
  local search_element="$2"
  for element in "${array[@]}"; do
    if [[ "${element}" == ${search_element} ]]; then
      echo "true"
      return
    fi
  done
  echo "false"
}

function is_alert_fired() {
  local alert_name="$1"
  local filter_value="$2"
  local filter_key="$3"
  local filtered_result
  cluster_ip=$(kubectl get svc rancher-monitoring-prometheus -n cattle-monitoring-system -o jsonpath='{.spec.clusterIP}')
  found="false"
  for i in {1..10}; do 
    active_alerts=$(curl "http://${cluster_ip}:9090/api/v1/alerts")
    if [[ -z "${filter_key}" && -z "${filter_value}" ]]; then
      filtered_result="$(jq -c '.data.alerts[] | select( (.labels.alertname == '\"${alert_name}\"') and (.state == "firing"))' <<< $active_alerts)"
    else
      filtered_result="$(jq -c '.data.alerts[] | select( (.labels.alertname == '\"${alert_name}\"') and select('$filter_key' == '\"$filter_value\"') and (.state == "firing"))' <<< $active_alerts)"
    fi
    
    if [[ -n "${filtered_result}" ]]; then
      found="true"
      break
    else
      sleep 60
    fi
  done
  echo "${found}"
}

export KUBECONFIG=/etc/rancher/rke2/rke2.yaml PATH=$PATH:/var/lib/rancher/rke2/bin

function mark_test_failed() {
  tmp=$(mktemp)
  jq --arg testid "$1" '. += [{"test" : $testid, "result": false}]' "$2" > "$tmp" && mv "$tmp" "$2"
}

function mark_test_passed() {
  tmp=$(mktemp)
  jq --arg testid "$1" '. += [{"test" : $testid, "result": true}]' "$2" > "$tmp" && mv "$tmp" "$2"
}

function create_certs() {
  local folder="$1"
  local dns_list="$2"
  local counter=1
  IFS=',' read -a dnsArray <<< "$dns_list"
  mkdir -p "${folder}"
  openssl genrsa -des3 -out "${folder}/rootCA.key" -passout pass:12345 2048
  openssl req -x509 -new -nodes -key "${folder}/rootCA.key" --passin pass:12345 -sha256 -days 1024  -out "${folder}/rootCA.crt" -subj "/C=US/ST=NY/O=UiPath, Inc./CN=UiPath, Service Fabric"
  
  echo -e "authorityKeyIdentifier=keyid,issuer\nbasicConstraints=CA:FALSE\nkeyUsage = digitalSignature, nonRepudiation, keyEncipherment, dataEncipherment\nextendedKeyUsage = serverAuth\nsubjectAltName = @alt_names\n[alt_names]" > "${folder}/v3.ext"

  for i in "${dnsArray[@]}"
  do
    printf "%s\n" "DNS.$counter = $i" >> "${folder}/v3.ext"
    counter=$((counter+1))
  done

  openssl req -new -nodes -out "${folder}/server.csr" -newkey rsa:2048 -keyout "${folder}/server.key" -subj "/C=US/ST=NY/O=UiPath, Inc./CN=Service Fabric"
  openssl x509 -req -in "${folder}/server.csr" -CA "${folder}/rootCA.crt" -CAkey "${folder}/rootCA.key"  --passin pass:12345  -CAcreateserial -out "${folder}/server.crt" -days 500 -sha256 -extfile "${folder}/v3.ext"
  openssl pkcs12 -export -out "${folder}/identity.pfx" -inkey "${folder}/rootCA.key" -in "${folder}/rootCA.crt" --passin pass:12345 -passout pass:12345
}