#!/bin/bash
# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/upgrade_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/longhornutils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

export OFFLINE_TMP_FOLDER="/tmp"
export HELM_EXPERIMENTAL_OCI=1

function get_latest_engine_image() {
  local version=$1
  local engine_images
  local image
  local try=0
  local maxtry=10
  while (( try != maxtry )) ; do
    engine_images=$(curl "${LONGHORN_URL}/v1/engineimages?") || true
    if [[ -n "${engine_images}" && -n $(echo "${engine_images}"|jq -r ".data[]") ]]; then
      image=$(echo "${engine_images}"|jq -r ".data[]|select(.version == \"${version}\").image")
      if [[ -n "${image}" ]]; then
        export LATEST_ENGINE_IMAGE="${image}"
        break;
      fi
    fi
    try=$((try+1))
    info "waiting for getting upgraded engine image details...${try}/${maxtry}"; sleep 30
  done
}

function get_current_pvc_engine() {
  local pv_name=$1
  local volume
  local engine_image

  volume=$(curl "${LONGHORN_URL}/v1/volumes/${pv_name}?") || true
  engine_image=$(echo "${volume}"|jq -r ".currentImage")

  export CURRENT_PVC_ENGINE="${engine_image}"
}

function upgrade_pvc_engine() {
  local pv_name=$1
  local engine_image=$2

  success=0
  if [[ -z "${engine_image}" ]]; then
    return
  fi

  local upgrade_resp
  upgrade_resp=$(curl "${LONGHORN_URL}/v1/volumes/${pv_name}?action=engineUpgrade"  --data-raw "{\"image\":\"${LATEST_ENGINE_IMAGE}\"}") || true
  if [[ -n "${upgrade_resp}" && "${upgrade_resp}" != "null" && -n $(echo "${upgrade_resp}"| jq -c ".id") ]]; then
    info "Upgrade Response: ${upgrade_resp}"
    wait_volume_healthy "${pv_name}"
    if [[ "${WAIT_VOLUME_HEALTHY}" -eq 1 ]]; then
      success=1
      unset WAIT_VOLUME_HEALTHY
    fi
  fi

  export UPGRADE_PVC_ENGINE_RESP=${success}
}

function upgrade_engine_image_for_pvc_attached_to_deployments() {
  local deployment_list

  # list of all deployment using PVC
  deployment_list=$(kubectl get deployments -A -o=json | jq -c ".items[] |\
{name: .metadata.name, namespace: .metadata.namespace, replicas: .spec.replicas, pvcList:.spec.template.spec |\
[select( has (\"volumes\") ).volumes[] |\
select( has (\"persistentVolumeClaim\") ).persistentVolumeClaim.claimName]|select(length > 0) }")

  for deployment in ${deployment_list};
  do
    deployment_name=$(echo "${deployment}"| jq -r ".name")
    pvc_list=$(echo "${deployment}"| jq -r ".pvcList[]")
    namespace=$(echo "${deployment}"| jq -r ".namespace")
    replica=$(echo "${deployment}" | jq -r ".replicas")
    local upgrade_required=0

    for pvc_name in ${pvc_list};
    do
      # get pv for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")

      get_current_pvc_engine "${pv_name}"
      if [[ "${LATEST_ENGINE_IMAGE}" == "${CURRENT_PVC_ENGINE}" ]]; then
        info "PVC volume already upgraded to new engine image ${CURRENT_PVC_ENGINE}"
        continue
      fi

      upgrade_required=1
    done

    if [ "${upgrade_required}" == 0 ]; then
      continue
    fi

    info "deployment $deployment_name is having issue"
    scale_down_deployment "${deployment_name}" "${namespace}"

    for pvc_name in ${pvc_list};
    do
      # get pv for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")

      upgrade_pvc_engine "${pv_name}" "${CURRENT_PVC_ENGINE}"

      if [[ "${UPGRADE_PVC_ENGINE_RESP}" -eq 0 ]]; then
        scale_up_deployment "${deployment_name}" "${namespace}" "${replica}"
        error "Upgrade failed for pvc ${pvc_name}, rollback deployment ${deployment_name}."
      fi
      unset UPGRADE_PVC_ENGINE_RESP
    info "Successfully upgraded engine image for pvc ${pvc_name}"
    done

    unset CURRENT_PVC_ENGINE
    scale_up_deployment "${deployment_name}" "${namespace}" "${replica}"
    sleep 5
  done
}

function upgrade_engine_image_for_pvc_attached_to_statefulsets() {
    local statefulset_list

  # list of all statefulset using PVC
  statefulset_list=$(kubectl get statefulset -A -o=json | jq -c ".items[] |\
{name: .metadata.name, namespace: .metadata.namespace, replicas: .spec.replicas, claimName:.spec |\
select( has (\"volumeClaimTemplates\") ).volumeClaimTemplates[].metadata.name, ownerReferences: .metadata.ownerReferences }")

  info "StatefulSet list with pvc attached: ${statefulset_list}"
  for statefulset in ${statefulset_list};
  do
    local statefulset_name
    local claimtemplate_name
    local namespace
    local replica
    local ownerReferences
    statefulset_name=$(echo "${statefulset}"| jq -r ".name")
    claimtemplate_name=$(echo "${statefulset}"| jq -r ".claimName")
    namespace=$(echo "${statefulset}"| jq -r ".namespace")
    replica=$(echo "${statefulset}" | jq -r ".replicas")
    ownerReferences=$(echo "${statefulset}" | jq -c ".ownerReferences")

    local pvc_prefix upgrade_required=0
    pvc_prefix="${claimtemplate_name}-${statefulset_name}"

    for((i=0;i<"${replica}";i++))
    do
      local pvc_name
      pvc_name="${pvc_prefix}-${i}"

      # get pv for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")

      get_current_pvc_engine "${pv_name}"
      if [[ "${LATEST_ENGINE_IMAGE}" == "${CURRENT_PVC_ENGINE}" ]]; then
        info "PVC volume already upgraded to new engine image ${CURRENT_PVC_ENGINE}"
        continue
      fi

      upgrade_required=1
      break
    done

    if [ "${upgrade_required}" == 0 ]; then
      continue
    fi

    info "Scaling down Statefulset ${statefulset_name} with ${replica} under namespace ${namespace}"
    # In case of rabbitmq and mongodb, we need to also scale down relevant operator
    if [[ ${statefulset_name} == "rabbitmq-server" && ${namespace} == "rabbitmq" ]]; then
      scale_down_deployment "rabbitmq-cluster-operator" "rabbitmq-system"
    elif [[ ${statefulset_name} == "mongodb-replica-set" && ${namespace} == "mongodb" ]]; then
      scale_down_deployment "mongodb-kubernetes-operator" "mongodb"
    fi
    scale_down_statefulset "${statefulset_name}" "${namespace}" "${ownerReferences}"

    for((i=0;i<"${replica}";i++))
    do
      local pvc_name
      pvc_name="${pvc_prefix}-${i}"

      # get pv for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")

      get_current_pvc_engine "${pv_name}"
      if [[ "${LATEST_ENGINE_IMAGE}" == "${CURRENT_PVC_ENGINE}" ]]; then
        info "PVC volume already upgraded to new engine image ${CURRENT_PVC_ENGINE}"
        continue
      fi

      upgrade_pvc_engine "${pv_name}" "${CURRENT_PVC_ENGINE}"
      unset CURRENT_PVC_ENGINE

      if [[ "${UPGRADE_PVC_ENGINE_RESP}" -eq 0 ]]; then
        if [[ ${statefulset_name} == "rabbitmq-server" && ${namespace} == "rabbitmq" ]]; then
          scale_up_deployment "rabbitmq-cluster-operator" "rabbitmq-system" "1"
        elif [[ ${statefulset_name} == "mongodb-replica-set" && ${namespace} == "mongodb" ]]; then
          scale_up_deployment "mongodb-kubernetes-operator" "mongodb" "1"
        fi
        scale_up_statefulset "${statefulset_name}" "${namespace}" "${replica}" "${ownerReferences}"
        error "Upgrade failed for pvc ${pvc_name}, rollback statefulset ${statefulset_name}."
      fi
      unset UPGRADE_PVC_ENGINE_RESP
    info "Successfully upgraded engine image for pvc ${pvc_name}"
    done

    if [[ ${statefulset_name} == "rabbitmq-server" && ${namespace} == "rabbitmq" ]]; then
      scale_up_deployment "rabbitmq-cluster-operator" "rabbitmq-system" "1"
    elif [[ ${statefulset_name} == "mongodb-replica-set" && ${namespace} == "mongodb" ]]; then
      scale_up_deployment "mongodb-kubernetes-operator" "mongodb" "1"
    fi
    scale_up_statefulset "${statefulset_name}" "${namespace}" "${replica}" "${ownerReferences}"
    sleep 5
  done
}

function get_pvc_resources() {

  get_pvc_resources_resp=""

  local PVC_NAME=$1
  local PVC_NAMESPACE=$2

  # PV have one to one mapping with PVC
  PV_NAME=$(kubectl -n "${PVC_NAMESPACE}" get pvc "${PVC_NAME}" -o json|jq -r ".spec.volumeName")

  get_pvc_resources_resp="{\"pvc_name\": \"${PVC_NAME}\", \"pv_name\": \"${PV_NAME}\"}"
  echo "${get_pvc_resources_resp}"
}

function upgrade_longhorn() {
    # Manually upgrading engine image
    export LONGHORN_COMPONENT_VERSION="v${COMPONENT_TARGET_VERSION}"
    # component version v1.1.2 -> longhorn version -> 1.1.2
    export LONGHORN_VERSION="${COMPONENT_TARGET_VERSION:1}"

    # set longhorn url
    set_longhorn_url

    # check longhorn reachable
    validate_longhorn_reachable

    # get latest engine image
    get_latest_engine_image "${LONGHORN_COMPONENT_VERSION}"

    if [[ -z "${LATEST_ENGINE_IMAGE}" ]]; then
      error "No engine image found for version: ${LONGHORN_COMPONENT_VERSION}"
    fi
    info "Upgrading PVC volumes to version: ${LONGHORN_COMPONENT_VERSION} with engine image: ${LATEST_ENGINE_IMAGE}"

    # upgrade longhorn
    upgrade_engine_image_for_pvc_attached_to_deployments
    upgrade_engine_image_for_pvc_attached_to_statefulsets

    info "Upgrading Longhorn version ${LONGHORN_COMPONENT_VERSION} completed successfully."
}

function validate_pvc_ready_to_upgrade() {
  for _ in {1..10}; do
    pvc_unhealthy=$(kubectl -n longhorn-system get volumes.longhorn.io -o json| jq -r ".items[]| {name: .metadata.name, state: .status| select( .state != \"attached\" and  .state != \"detached\").state}")
    if [[ -z "${pvc_unhealthy}" ]]; then
      info "All pvc in healthy state for engineUpgrade"
      return
    fi
    info "Few pvc's are in unhealthy state, PVC List:  ${pvc_unhealthy}. Will retry.."
    sleep 60
  done
  error "Few pvc's are in unhealthy state, PVC List:  ${pvc_unhealthy}, please verify and then run longhorn upgrade."
}

function validate_longhorn_engine_version()  {
  COMPONENT_TARGET_VERSION="$(jq --arg component "longhorn" -r '.vendor_charts[] | select(.name==$component) | .version' < "${TARGET_VERSION_FILE_PATH}")"
  COMPONENT_CURRENT_VERSION="$(get_version_for_component "${COMPONENT}")"
  if [ "${COMPONENT_CURRENT_VERSION}" = "${COMPONENT_TARGET_VERSION}" ]; then
    local engine_name
    engine_name=$(kubectl -n longhorn-system get volumes.longhorn.io -o json|jq -r ".items[].spec| select ( .engineImage | contains(\"${COMPONENT_TARGET_VERSION}\" ) | not ).engineImage")
    if [[ -z "${engine_name}" ]]; then
      info "Component ${COMPONENT} already running at ${COMPONENT_TARGET_VERSION}. Skipping step"
      exit 0
    fi
  fi
}

verify_longhorn_engines_upgraded() {
  ready=0
  image_tag=$(kubectl get engineimages -n longhorn-system --sort-by=.metadata.creationTimestamp -o 'jsonpath={.items[-1].status.version}')

  for _ in {1..10}; do
    local volumes engines
    volumes=$(kubectl -n longhorn-system get volumes.longhorn.io -o json|jq -r ".items[].spec| select ( .engineImage | contains(\"${image_tag}\" ) | not ).engineImage")
    engines=$(kubectl -n longhorn-system get engines.longhorn.io -o json|jq -r ".items[].spec| select ( .engineImage | contains(\"${image_tag}\" ) | not ).engineImage")
    if [[ -z "${volumes}" && -z "${engines}" ]]; then
      ready=1 && break
    fi
    sleep 10
    volumes=$(kubectl -n longhorn-system get engines.longhorn.io -o json|jq -r ".items[].spec| select ( .engineImage | contains(\"${image_tag}\" ) | not ).volumeName")
    info "${volumes} are not upgraded yet"
  done
  [[ ${ready} -ne 1 ]] && warn "Longhorn volumes are not upgraded yet" && return 1
  info "Longhorn volumes are upgraded" && return 0
}

#MAIN
function main() {
  if [[ "${ETCD_RESTORE}" == "true" ]]; then
    return
  fi

  if [[ "${NODE_ROLE}" == "none" || "${NODE_ROLE}" == "server" ]]; then
    local volumes engines
    image_tag=$(kubectl get engineimages -n longhorn-system --sort-by=.metadata.creationTimestamp -o 'jsonpath={.items[-1].status.version}')
    volumes=$(kubectl -n longhorn-system get volumes.longhorn.io -o json|jq -r ".items[].spec| select ( .engineImage | contains(\"${image_tag}\" ) | not ).engineImage")
    engines=$(kubectl -n longhorn-system get engines.longhorn.io -o json|jq -r ".items[].spec| select ( .engineImage | contains(\"${image_tag}\" ) | not ).engineImage")
    if [[ -z "${volumes}" && -z "${engines}" ]]; then
      info "Longhorn engine images are already upgraded"
      return
    fi

    # validate longhorn step
    validate_longhorn_engine_version

    # validate all pvc in health or detached state
    validate_pvc_ready_to_upgrade

    # longhorn upgrade
    upgrade_longhorn

    verify_longhorn_engines_upgraded || error "Longhorn engine images are not upgraded. Please Re-run upgrade"
  fi
}

export STEP=$1
export COMPONENT="longhorn"
# Pass additional arguments from upgrade path to main function.
main "${@:2}"
