#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu
set -o pipefail

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
UPGRADE_PATHS_DIR=$(cd "${SCRIPT_DIR}/paths" && pwd)
if [[ -z "${UPGRADE_PATHS_DIR}" ]]; then
  error "Could not determine upgrade paths path"
fi
UPGRADE_VENDOR_DIR=$(cd "${SCRIPT_DIR}/vendor" && pwd)
if [[ -z "${UPGRADE_VENDOR_DIR}" ]]; then
  error "Could not determine vendor path"
fi
SEMVER="${MODULES_DIR}/semver2.sh"
SEMVER_CUSTOM="${MODULES_DIR}/semver-custom.sh"

declare -a DEBUG_FLAG
declare -a DEBUG_ARGS

export SEMVER
export SEMVER_CUSTOM

readonly SCRIPT_DIR
readonly MODULES_DIR
readonly UPGRADE_VENDOR_DIR
export UPGRADE_VENDOR_DIR

export DEFAULTSFILE="${SCRIPT_DIR}/defaults.json"
export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_input.json"
readonly VERSION_FILE_PATH="${SCRIPT_DIR}/versions.json"
readonly TARGET_VERSION_FILE_PATH="${SCRIPT_DIR}/../versions/helm-charts.json"
export TARGET_VERSION_FILE_PATH
export VERSION_FILE_PATH
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/upgrade_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/cluster_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/cilium_utils.sh"

export NODE_ROLE=${NODE_ROLE:-none}

#shellcheck disable=SC2034
UPGRADE_STEPS=()

function parse_input_file() {
  parse_fqdn_values
  VERSION="$(get_mandatory_json_value ".version")"
  PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
  DOCKER_REGISTRY_URL="$(get_mandatory_json_value ".registries.docker.url")"
  DOCKER_REGISTRY_NODEPORT="$(get_mandatory_json_value ".infra.docker_registry.node_port")"
  PROFILE="$(get_mandatory_json_value ".profile")"
  #shellcheck disable=SC2034
  RKE_TOKEN="$(get_mandatory_json_value ".rke_token")" # used in the telemetry checkpoint

  export VERSION
  export PROXY_ENABLED
  export DOCKER_REGISTRY_URL
  export PROFILE
}

function parse_upgrade_args() {
  INFILE=""
  OUTFILE=""

  while getopts 'i:o:k:f:s:c:a:d' flag "$@"; do
    case "${flag}" in
    d)
      info "Running in debug"
      set -x
      DEBUG_ARGS+=("-x")
      DEBUG_FLAG+=("-d")
      ;;
    i)
      info "Using input parameters file: ${OPTARG}"
      INFILE=${OPTARG}
      ;;
    c)
      UPGRADECNI=${OPTARG}
      ;;
    k)
      UPGRADEINFRA=${OPTARG}
      ;;
    f)
      UPGRADEFABRIC=${OPTARG}
      ;;
    s)
      UPGRADESERVICES=${OPTARG}
      ;;
    a)
      INSTALLSTANDALONEAICENTER=${OPTARG}
      ;;
    o)
      info "Using output parameters file: ${OPTARG}"
      OUTFILE=${OPTARG}
      ;;
    :)
      error "Invalid option: ${OPTARG} requires an argument."
      ;;
    *)
      error "Unexpected option ${flag}"
      ;;
    esac
  done
  if [[ -z "${INFILE}" || -z "${OUTFILE}" || -z "${UPGRADEINFRA}" || -z "${UPGRADEFABRIC}" || -z "${UPGRADESERVICES}" || -z "${UPGRADECNI}" ]]; then
    display_usage
    error "Mandatory parameters not found"
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error "File ${INFILE} not found"
  fi
  shift $((OPTIND - 1))
}

function determine_supported_upgrade() {
  CURRENT_VERSION=$(jq -r '."uipath"' "${VERSION_FILE_PATH}")
  servers_need_upgrading=$(jq -cr --arg VERSION "${VERSION}" '.nodes[] | select(.nodeType=="server" and .infraVersion!=$VERSION) | .name' "${VERSION_FILE_PATH}")
  nodes_need_upgrading=$(jq -cr --arg VERSION "${VERSION}" '.nodes[] | select(.infraVersion!=$VERSION) | .name' "${VERSION_FILE_PATH}")
  # get current's server version
  if [[ "${UPGRADEINFRA}" == "1" ]]; then
    # if agent, check if all servers were updated first
    if [[ "${NODE_ROLE}" == "agent" ]]; then
      #shellcheck disable=SC2016
      [[ -n ${servers_need_upgrading} ]] && error "$(echo "Must upgrade all servers before upgrading agents. Servers not upgraded: ${servers_need_upgrading}" | sed -e 'H;${x;s/\n/,/g;s/^,//;p;};d')"
      CURRENT_VERSION=$(kubectl get nodes "$(fetch_hostname_lowercase)" -o jsonpath='{.metadata.annotations.uipath/infra-version}')  
    fi
  fi
  if [[ "${UPGRADEFABRIC}" == "1" || "${UPGRADESERVICES}" == "1" ]]; then
    # check that current node is not an agent if we're gonna update fabric and services
    [[ "${NODE_ROLE}" == "agent" ]] && error_with_code "11029"
    # get number of nodes; skip nodes_need_upgrading check if single node environment
    nodes_count=$(jq -r '.nodes | length' "${VERSION_FILE_PATH}")
    [[ -n ${servers_need_upgrading} ]] && [[ "${UPGRADEINFRA}" != "1" ]] && (( nodes_count == 1 )) && error "Must upgrade the node's infrastructure. Rerun the command, adding the '-k' parameter."
    #shellcheck disable=SC2016
    [[ -n ${nodes_need_upgrading} ]] && (( nodes_count > 1 )) && error "$(echo "Must upgrade all nodes before upgrading fabric and services. Nodes not upgraded: ${nodes_need_upgrading}" | sed -e 'H;${x;s/\n/,/g;s/^,//;p;};d')"
  fi

  CAN_UPGRADE="false"
  current_apps_versions=$(jq -r '. | to_entries | .[] | select(.key!="nodes") | .key + "\t" + .value' < "${VERSION_FILE_PATH}")
  while read -r COMPONENT CURRENT_COMPONENT_VERSION; do
    TARGET_COMPONENT_VERSION=$(get_target_version_for_component "${COMPONENT}")
    [ "$(${SEMVER_CUSTOM} "${CURRENT_COMPONENT_VERSION}" "${TARGET_COMPONENT_VERSION}")" != 0 ] && CAN_UPGRADE="true"
  done <<< "${current_apps_versions}"
  # Validate the installed version is older than the target version
  if [ "$(${SEMVER_CUSTOM} "${CURRENT_VERSION}" "${VERSION}")" != -1 ]; then
    info "Current version: ${CURRENT_VERSION}"
    info "Upgrading to version: ${VERSION}"
    if [ "${CAN_UPGRADE}" == "true" ]; then
      info "Some components were not fully upgraded. Retrying upgrade."
    else
      error "You are already running a newer version. Exiting upgrade."
    fi
  fi

  # Determine type of upgrade
  local suite_type
  suite_type="full"
  [ "${INSTALLSTANDALONEAICENTER}" == "1" ] && suite_type="aicenter"

  for file in "${UPGRADE_PATHS_DIR}"/*.json; do
    TEMP_INFILE="${INFILE}"
    INFILE="${file}"
    # Get supported source and target versions from upgrade path file
    MIN_SOURCE_VERSION="$(get_mandatory_json_value ".minSourceVersion")"
    MAX_SOURCE_VERSION="$(get_mandatory_json_value ".maxSourceVersion")"
    MIN_TARGET_VERSION="$(get_mandatory_json_value ".minTargetVersion")"
    MAX_TARGET_VERSION="$(get_mandatory_json_value ".maxTargetVersion")"
    PATH_SUITE_TYPE="$(get_mandatory_json_value ".type")"

    # VERSION should be >=MIN_TARGET_VERSION & <=MAX_TARGET_VERSION
    # CURRENT_VERSION should be >=MIN_SOURCE_VERSION & <=MAX_SOURCE_VERSION
    if [ "${PATH_SUITE_TYPE}" == "${suite_type}" ] && [ "$(${SEMVER_CUSTOM} "${VERSION}" "${MIN_TARGET_VERSION}")" != -1 ] && [ "$(${SEMVER_CUSTOM} "${VERSION}" "${MAX_TARGET_VERSION}")" != 1 ] && [ "$(${SEMVER_CUSTOM} "${CURRENT_VERSION}" "${MIN_SOURCE_VERSION}")" != -1 ] && [ "$(${SEMVER_CUSTOM} "${CURRENT_VERSION}" "${MAX_SOURCE_VERSION}")" != 1 ]; then
      export upgrade_path_file=$file
    fi
    export INFILE="${TEMP_INFILE}"
  done
  if [[ -z "${upgrade_path_file+x}" ]]; then
    error "Upgrading from ${CURRENT_VERSION} to ${VERSION} is not supported"
  fi
}

function parse_upgrade_path() {
  local TEMP_INFILE="${INFILE}"
  export INFILE="${upgrade_path_file}"
  RAW_PRE_INFRA_STEPS="$(get_mandatory_json_value ".pre_infra_steps")"
  for row in $(echo "${RAW_PRE_INFRA_STEPS}" | jq -c '.[]'); do
    PRE_INFRA_UPGRADE_STEPS+=("${row}")
  done
  RAW_POST_INFRA_STEPS="$(get_mandatory_json_value ".post_infra_steps")"
  for row in $(echo "${RAW_POST_INFRA_STEPS}" | jq -c '.[]'); do
    POST_INFRA_UPGRADE_STEPS+=("${row}")
  done
  RAW_PRE_FABRIC_STEPS="$(get_mandatory_json_value ".pre_fabric_steps")"
  for row in $(echo "${RAW_PRE_FABRIC_STEPS}" | jq -c '.[]'); do
    PRE_FABRIC_UPGRADE_STEPS+=("${row}")
  done
  RAW_POST_FABRIC_STEPS="$(get_mandatory_json_value ".post_fabric_steps")"
  for row in $(echo "${RAW_POST_FABRIC_STEPS}" | jq -c '.[]'); do
    POST_FABRIC_UPGRADE_STEPS+=("${row}")
  done
  RAW_PRE_SERVICES_STEPS="$(get_mandatory_json_value ".pre_services_steps")"
  for row in $(echo "${RAW_PRE_SERVICES_STEPS}" | jq -c '.[]'); do
    PRE_SERVICES_UPGRADE_STEPS+=("${row}")
  done
  RAW_POST_SERVICES_STEPS="$(get_mandatory_json_value ".post_services_steps")"
  for row in $(echo "${RAW_POST_SERVICES_STEPS}" | jq -c '.[]'); do
    POST_SERVICES_UPGRADE_STEPS+=("${row}")
  done
  INFILE="${TEMP_INFILE}"
}

function execute_upgrade_step() {
  local STEP
  STEP=$1
  SCRIPT=$(echo "${STEP}" | jq -r '.upgradeScriptName')
  ARGS=$(echo "${STEP}" | jq -r '.upgradeScriptArguments')
  /bin/bash "${DEBUG_ARGS[@]}" "${SCRIPT_DIR}/scripts/${SCRIPT}" "${STEP}" "${ARGS}"
}

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"

########
# Get version of running cluster
########
function init_version_file() {
  touch "${VERSION_FILE_PATH}"
  echo "{}" >"${VERSION_FILE_PATH}"
}

function get_current_version() {
  get_rke2_versions
  if [[ "${NODE_ROLE}" != "agent" ]]; then
    if [[ ${INSTALL_TYPE} == "offline" ]]; then
      get_docker_registry_version
    fi
    get_longhorn_version
    get_argocd_version
    get_apps_versions
  fi
}

function configure_install_type() {
  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    export DOCKER_REGISTRY_URL="localhost:${DOCKER_REGISTRY_NODEPORT}"
  fi
}

function get_service_installation_type() {
  SERVICE_INSTALL_TYPE=$(get_cluster_config_default "SERVICE_INSTALL_TYPE" "all")
  info "Service installation type: ${SERVICE_INSTALL_TYPE}"
}

function init_argocd() {
  local BASE_DIR
  BASE_DIR=$(cd "${SCRIPT_DIR}/../Fabric_Installer" && pwd)
  get_fabric_install_paths "${BASE_DIR}"
  get_argocd_install_params
  login_argocd "${ARGOCD_NODEPORT_HTTP}" "${ARGOCD_FQDN}" "${ARGOCD_PASSWORD_SECRET_NAME}" "${ARGOCD_NAMESPACE}" "argocd"
  disable_maintenance_mode --no-argo-login
}

function main() {
  cd "${SCRIPT_DIR}"
  init_version_file
  parse_upgrade_args "$@"
  parse_input_file
  configure_install_type

  if [[ "${UPGRADECNI}" == "1" ]]; then
    if [[ "$(is_cilium_installed)" == "true" ]]; then
      info "Cilium is already installed."
      exit 0
    fi
    [[ "${NODE_ROLE}" == "none" || "${NODE_ROLE}" == "server" ]] && {
      install_cilium
      cleanup_longhorn_networkpolicy
    }
    warn "This node needs to be rebooted (OS restart)."
    exit 0
  fi

  get_current_version
  determine_supported_upgrade
  parse_upgrade_path
  check_available_masters

  get_service_installation_type

  if [[ "${UPGRADEINFRA}" == "1" ]]; then
    for step in "${PRE_INFRA_UPGRADE_STEPS[@]}"; do
      execute_upgrade_step "${step}"
    done

    "${SCRIPT_DIR}"/../Infra_Installer/infra-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/infra_upgrade_output.json" "${DEBUG_FLAG[@]}"

    for step in "${POST_INFRA_UPGRADE_STEPS[@]}"; do
      execute_upgrade_step "${step}"
    done

    # Mark this node as updated
    update_node_annotation_value "$(fetch_hostname_lowercase)" "uipath/infra-version" "${VERSION}"
  fi

  if [[ "${UPGRADEFABRIC}" == "1" || "${UPGRADESERVICES}" == "1" ]]; then
    init_argocd
  fi

  if [[ "${UPGRADEFABRIC}" == "1" ]]; then

    for step in "${PRE_FABRIC_UPGRADE_STEPS[@]}"; do
      execute_upgrade_step "${step}"
    done

    "${SCRIPT_DIR}"/../Fabric_Installer/fabric-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/fabric_upgrade_output.json" "${DEBUG_FLAG[@]}"

    for step in "${POST_FABRIC_UPGRADE_STEPS[@]}"; do
      execute_upgrade_step "${step}"
    done
  fi

  if [[ "${UPGRADESERVICES}" == "1" ]]; then
    for step in "${PRE_SERVICES_UPGRADE_STEPS[@]}"; do
      execute_upgrade_step "${step}"
    done

    if [[ "${SERVICE_INSTALL_TYPE}" == "all" ]]; then
      # Execute Services Upgrade
      "${SCRIPT_DIR}"/../Services_Installer/services-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/services_upgrade_output.json" "${DEBUG_FLAG[@]}"
    else
      # Execute Standalone Aicenter Services Upgrade
      "${SCRIPT_DIR}"/../Services_Installer/aicenter-service-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/services_upgrade_output.json" "${DEBUG_FLAG[@]}"
    fi

    for step in "${POST_SERVICES_UPGRADE_STEPS[@]}"; do
      execute_upgrade_step "${step}"
    done
  fi

  # Put the whole upgrade output in following file
  UPGRADE_OUTFILE="${SCRIPT_DIR}/../upgrade_output.json"
  echo "{}" >"${UPGRADE_OUTFILE}"
  if [[ -f "${SCRIPT_DIR}/infra_upgrade_output.json" ]]; then
    jq -s '.[0] * .[1]' "${UPGRADE_OUTFILE}" "${SCRIPT_DIR}/infra_upgrade_output.json" >"tempout.json"
    cat "tempout.json" >"${UPGRADE_OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/fabric_upgrade_output.json" ]]; then
    jq -s '.[0] * .[1]' "${UPGRADE_OUTFILE}" "${SCRIPT_DIR}/fabric_upgrade_output.json" >"tempout.json"
    cat "tempout.json" >"${UPGRADE_OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/services_upgrade_output.json" ]]; then
    jq -s '.[0] * .[1]' "${UPGRADE_OUTFILE}" "${SCRIPT_DIR}/services_upgrade_output.json" >"tempout.json"
    cat "tempout.json" >"${UPGRADE_OUTFILE}"
    rm "tempout.json"
  fi
  
  if [[ "${UPGRADESERVICES}" == "1" ]]; then
    read_cluster_id
    send_checkpoint_telemetry "Upgrade: Success"
  fi

}

main "$@"
