#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error_with_code "10001"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error_with_code "10002"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/bin:/usr/local/bin
export HELM_EXPERIMENTAL_OCI=1
export OFFLINE_TMP_FOLDER="/tmp"
export DOT_FOLDER=~/.uipath

# Set logfile path if not already set by caller
: "${INSTALLER_LOGFILE:="${SCRIPT_DIR}/install-$(date +'%Y-%m-%dT%H_%M_%S').log"}"

# Redirect /dev/null to INSTALLER_LOGFILE if set
dev_null="/dev/null"
[[ -n "${INSTALLER_LOGFILE:-}" ]] && dev_null="${INSTALLER_LOGFILE}"

# Check if file descriptor 3 should be declared here or inherited from parent script
# Enables us to redirect stdout to a file but still print some special output to the terminal (like user prompts)
# Example: ./install-uipath.sh 3>&1 >/tmp/install.log
export FD3_OUTPUT=false
if { true >&3; } 2> "${dev_null}"; then
  FD3_OUTPUT=true
fi
$FD3_OUTPUT || { echo "File descriptor 3 not found. Opening it now ..." >> "${dev_null}"; exec 3>&1; }

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/offline-utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/install_packages/prerequisites-rhel.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/upgrade_utils.sh

export DEFAULTSFILE="${SCRIPT_DIR}/defaults.json"
export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_input.json"

export ETCD_RESTORE=${ETCD_RESTORE:-false}

export UPGRADE_SILENT_MODE=${UPGRADE_SILENT_MODE:-false}

INSTALL_ID="$(uuidgen)"
export INSTALL_ID
readonly INSTALL_ID

COMPARE_CONFIG=1
COMPARE_CONFIG_ONLY=0
ORIGINAL_INFILE=""

INSTALLCNI=0
INSTALLINFRA=0
INSTALLFABRIC=0
INSTALLSERVICES=0
# Exporting as it's required in infra-install.sh
export DO_UPGRADE=0
SETUPBACKUP=0
SETUPRESTORE=0
CHECKRERUN=0
INSTALLSTANDALONEAICENTER=0
INSTALL_OFFLINE_PREREQS=0
SKIP_PACKAGES_INSTALL=0
SKIP_PRE_REQS=0
INSTALL_TYPE=""
MACHINES=""
FORCEUPDATEFQDN=0

declare -a DEBUG_ARGS
declare -a DEBUG_FLAG

#FUNCTIONS
function display_usage() {
  echo "usage: $(basename "$0") [-hd] -i infile"
  echo "  -d                                                  Run in debug"
  echo "  -h                                                  Display help"
  echo "  -c                                                  Install CNI"
  echo "  -k|--install-kubernetes                             Install Kubernetes and infrastructure modules"
  echo "  -j|--join-node <server|agent|task-mining|gpu>       Type of node to be joined to an existing cluster"
  echo "  --kubeconfig                                        Specify path to the kubeconfig file"
  echo "  -f|--install-fabric                                 Install Fabric"
  echo "  -s|--install-services                               Install UiPath Services"
  echo "  -a|--install-all                                    Run all installers in order"
  echo "  -b|--backup                                         Setup Backup"
  echo "  -r|--restore                                        Setup Restore"
  echo "  --restore-volume                                    Restore Volume"
  echo "  --disable-backup                                    Disable Backup"
  echo "  -i|--input <input.json>                             Specify input json file"
  echo "  -o|--output <output.json>                           Specify output json file"
  echo "  --install-type <online|offline>                     Specify installation type. Defaults to online. Cannot be set to online together with --offline-bundle."
  echo "  --offline-bundle </tmp/offline_bundle.tar.gz>       Specify the path to the offline bundle"
  echo "  --offline-tmp-folder </uipath>                      Specify parent directory for the temporary offline install folder"
  echo "  --install-offline-prereqs                           Install offline prerequisites"
  echo "  -u|--upgrade                                        Upgrade an existing Automation Suite cluster"
  echo "  --skip-packages-install                             Disable installation of package dependencies(helm, jq, etc.)"
  echo "  --skip-pre-reqs                                     Skips pre-requisite checks during install"
  echo "  --install-standalone-aicenter                       Install standalone aicenter service only"
  echo "  --one-time-manual-backup                            Trigger Manual Backup and disable it after completion."
  echo "  -m|--machines <comma separated machine's IP>        Comma separated IP address / hostnames of all the machines"
  echo "  --compare-config                                    Only compare current config with the cluster state"
  echo "  --skip-compare-config                               Skip compare current config with the cluster state"
}

function contains() {
  if [[ $1 =~ (^|[[:space:]])$2($|[[:space:]]) ]]; then
    echo "0"
  else
    echo "1"
  fi
}

#######
# Parse long input arguments
# Either -a or --argument or --argument-with-dashes
# To pass a value to an argument, separate with space
# --argument value
######
function parse_long_args() {
  local PARAMS=""
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      DEBUG=1
      shift
      set -x
      DEBUG_ARGS+=("-x")
      DEBUG_FLAG+=("-d")
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -k | --install-kubernetes)
      INSTALLINFRA=1
      shift
      ;;

    -c | --install-cni)
      INSTALLCNI=1
      shift
      ;;

    -j | --join-node)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODE_TYPE="$2"
        if [[ "${NODE_TYPE}" == "server" ]]; then
          NODE_ROLE="server"
        else
          NODE_ROLE="agent"
        fi
        result=$(contains "server agent task-mining gpu" "${NODE_TYPE}")
        [[ "${result}" == "1" ]] && display_usage && error_with_code "10020" "${NODE_TYPE}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --kubeconfig)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        [[ -f "${2}" ]] || error_with_code "10006" "$2"
        KUBECONFIG_PATH="$2"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    -f | --install-fabric)
      INSTALLFABRIC=1
      shift
      ;;

    -s | --install-services)
      INSTALLSERVICES=1
      shift
      ;;

    -a | --install-all)
      INSTALLINFRA=1
      INSTALLFABRIC=1
      INSTALLSERVICES=1
      shift
      ;;

    -b|--backup)
      SETUPBACKUP=1
      shift
      ;;

    -r|--restore)
      SETUPRESTORE=1
      shift
      ;;

    --volume-restore)
      RESTORE_VOLUME="true"
      shift
      ;;

    --disable-backup)
      DISABLE_BACKUP="true"
      shift
      ;;

    -m|--machines)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        MACHINES="$2"
        shift 2
      else
        display_usage
        error_with_code "10005" "${1}"
        shift 2
      fi
      ;;

    --skip-pre-reqs)
      SKIP_PRE_REQS=1
      shift
      ;;

    -i | --input)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        INFILE=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    -o | --output)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OUTFILE=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --offline-bundle)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OFFLINE_BUNDLE_PATH=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --offline-tmp-folder)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OFFLINE_TMP_FOLDER=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;


    --install-offline-prereqs)
      INSTALL_OFFLINE_PREREQS=1
      shift
      ;;

    -u|--upgrade)
      export DO_UPGRADE=1
      shift
      ;;

    --accept-license-agreement)
      export LICENSE_AGREEMENT="accept"
      shift
      ;;

    --install-standalone-aicenter)
      INSTALLSTANDALONEAICENTER=1
      shift
      ;;

    --skip-packages-install)
      SKIP_PACKAGES_INSTALL=1
      shift
      ;;

    --non-interactive-upgrade)
      UPGRADE_SILENT_MODE=true
      shift
      ;;

    --one-time-manual-backup)
      ONE_TIME_MANUAL_BACKUP="true"
      SETUPBACKUP=1
      shift
      ;;

    --compare-config)
      COMPARE_CONFIG_ONLY=1
      shift
      ;;

    --skip-compare-config)
      COMPARE_CONFIG=0
      shift
      ;;

    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        INSTALL_TYPE="$2"
        [[ "${INSTALL_TYPE}" != "online" && "${INSTALL_TYPE}" != "offline" ]] && display_usage && error_with_code "10028" "${INSTALL_TYPE}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --force-update-fqdn)
      FORCEUPDATEFQDN=1
      shift
      ;;

    -*) # unsupported flags
      display_usage
      error_with_code "10007" "$1"
      exit 1
      ;;
    *) # preserve positional arguments
      PARAMS="$PARAMS $1"
      shift
      ;;
    esac
  done

  [[ "${INSTALL_TYPE}" == "online" ]] && [[ -n ${OFFLINE_BUNDLE_PATH} ]] && error_with_code "10029"
  [[ -n "${OFFLINE_BUNDLE_PATH}" ]] && INSTALL_TYPE="offline"
  [[ -z "${INSTALL_TYPE}" ]] && INSTALL_TYPE="online"
  [[ "${INSTALL_TYPE}" == "offline" ]] && [[ -z ${OFFLINE_BUNDLE_PATH} ]] && ((INSTALLINFRA == 1)) && error_with_code "10030"

  [[ "${INSTALL_OFFLINE_PREREQS}" == "1" ]] && [[ -z ${OFFLINE_BUNDLE_PATH} ]] &&
      error_with_code "10008"

  # if the offline bundle parameter exists:
  # - test if the file exists
  # - validate that the CRC matches the expected list
  if [[ -n "${OFFLINE_BUNDLE_PATH}" ]];
  then
    if [[ ! -f "${OFFLINE_BUNDLE_PATH}" ]];
    then
      error_with_code "10036" "${OFFLINE_BUNDLE_PATH}"
    fi
    if [[ "${INSTALLFABRIC}" == "1" || "${INSTALLSERVICES}" == "1" ]];
    then
      if ! check_crc "${OFFLINE_BUNDLE_PATH}" "sf";
      then
        error_with_code "10035" "${OFFLINE_BUNDLE_PATH}"
      fi
    else
      if ! check_crc "${OFFLINE_BUNDLE_PATH}" "sf" && ! check_crc "${OFFLINE_BUNDLE_PATH}" "sf-infra-only";
      then
        error_with_code "10035" "${OFFLINE_BUNDLE_PATH}"
      fi
    fi
    info "Bundle version matches the installer"
  fi

  # check if any installation will be executed
  if [[ "${INSTALLCNI}|${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}|${SETUPBACKUP}|${SETUPRESTORE}|${COMPARE_CONFIG_ONLY}" == "0|0|0|0|0|0|0" ]]; then
    info "At least one installer checks needs to be executed."
    display_usage
    exit 1
  fi

  # check -c is used alone
  if ((INSTALLINFRA + INSTALLFABRIC + INSTALLSERVICES + SETUPBACKUP + SETUPRESTORE + COMPARE_CONFIG_ONLY != 0)) && (( INSTALLCNI == 1 )); then
    info "-c must be use alone"
    display_usage
    exit 1
  fi

  # check --compare-config is used alone
  if ((INSTALLCNI + INSTALLINFRA + INSTALLFABRIC + INSTALLSERVICES + SETUPBACKUP + SETUPRESTORE != 0)) && (( COMPARE_CONFIG_ONLY == 1 )); then
    error_with_code "12014"
  fi

  # validate that fabric is not skipped
  if [[ "${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}" == "1|0|1" ]]; then
    error_with_code "12008"
  fi

  # don't install fabric and services when joining nodes
  if ((INSTALLFABRIC + INSTALLSERVICES != 0)) && [[ "${NODE_ROLE}" == "agent" ]]; then
    error_with_code "10010"
  fi

  # validate that no two backup & restore were selected
  if (( SETUPBACKUP + SETUPRESTORE == 2 )) ; then
    error_with_code "10012"
  fi

  # check if provided input file was provided and exists
  if [[ -z "${INFILE}" ]]; then
    error_with_code "10013"
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error_with_code "10014" "${INFILE}"
  fi
  # check if output file was provided
  if [[ -z "${OUTFILE}" ]]; then
    error_with_code "10015"
  fi

  # set positional arguments in their proper place
  eval set -- "$PARAMS"
}

########
# Parse input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_input_file() {
  VERSION="$(get_mandatory_json_value ".version")"
  #  while recovering cluster, we recovered fqdn from nfs server
  if (( SETUPRESTORE == 0 )); then
    parse_fqdn_values
  fi
  DOCKER_REGISTRY_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  # support for online private installation from OCI Helm registry
  HELM_REGISTRY_URL="$(get_mandatory_json_value ".registries.helm.url")"
  HELM_REGISTRY_USERNAME="$(get_optional_json_value ".registries.helm.username")"
  HELM_REGISTRY_PASSWORD="$(get_optional_json_value ".registries.helm.password")"
  PODMAN_FOLDER="$(get_mandatory_json_value ".infra.offline.dir")/lib/containers"
  ACTIVE_ACTIVE="$(get_optional_json_value ".active_active.enabled")"
  RKE_TOKEN="$(get_mandatory_json_value ".rke_token")"
  export HELM_REGISTRY_URL
  export HELM_REGISTRY_USERNAME
  export HELM_REGISTRY_PASSWORD

  # # infra specific versions
  DOCKER_REGISTRY_VERSION="$(get_chart_version_from_manifest "docker-registry")"
  SELF_HEAL_OPERATOR_VERSION="$(get_chart_version_from_manifest "self-heal-operator")"
  ISTIO_CONFIGURE_VERSION="$(get_chart_version_from_manifest "istio-configure")"
  RKE2_COREDNS_CONFIG_VERSION="$(get_chart_version_from_manifest "rke2-coredns-config")"
  LONGHORN_VERSION="$(get_chart_version_from_manifest "longhorn" "vendor")"
  RANCHER_ISTIO_VERSION="$(get_chart_version_from_manifest "rancher-istio" "vendor")"

  DOCKER_REGISTRY_NODEPORT="$(get_mandatory_json_value ".infra.docker_registry.node_port")"

  TELEMETRY_OPT_OUT="$(get_optional_json_value ".telemetry_optout")"

  PROFILE="$(get_mandatory_json_value ".profile")"
  if [[ "${PROFILE}" == "ha" ]];
  then
    INSTALL_MULTINODE="true"
  else
    INSTALL_MULTINODE="false"
  fi

  if [[ "${INSTALL_TYPE}" == "offline" ]];
  then
    TELEMETRY_OPT_OUT="true"
  fi

  export INSTALL_MULTINODE
  export TELEMETRY_OPT_OUT
  export CLUSTER_FQDN
  export INSTALLSTANDALONEAICENTER
  export VERSION
}

# setup local helm cache and charts for infra installation
# Helm charts are pulled from the OCI compliant registry and stored locally
# if HELM_REGISTRY_USERNAME and HELM_REGISTRY_PASSWORD are set, then the helm registry is private and authentication is required
function download_infra_charts() {
  info "Using local helm folder ${TMP_HELM_CHARTS_DIR}"
  if [[ -n "${HELM_REGISTRY_USERNAME}" ]] && [[ -n "${HELM_REGISTRY_PASSWORD}" ]]; then
    info "Using private Helm registry located at ${HELM_REGISTRY_URL}"
    echo "${HELM_REGISTRY_PASSWORD}" |
      helm registry login "${HELM_REGISTRY_URL}" \
        --username "${HELM_REGISTRY_USERNAME}" \
        --password-stdin &>> "${dev_null}"
  fi

  declare -a charts=(
    "helm/docker-registry ${DOCKER_REGISTRY_VERSION}"
    "vendor/longhorn ${LONGHORN_VERSION}"
    "vendor/rancher-istio ${RANCHER_ISTIO_VERSION}"
    "helm/istio-configure ${ISTIO_CONFIGURE_VERSION}"
    "helm/rke2-coredns-config ${RKE2_COREDNS_CONFIG_VERSION}"
    "helm/self-heal-operator ${SELF_HEAL_OPERATOR_VERSION}"
  )

  for crt_chart in "${charts[@]}"; do
    IFS=' ' read -ra crt_chart_arr <<<"${crt_chart}"

    local try=0
    local maxtry=10
    local status="notready"
    while [[ ${status} == "notready" ]] && ((try != maxtry)); do
      try=$((try + 1))
      info "Pulling chart ${HELM_REGISTRY_URL}/${crt_chart_arr[0]}:${crt_chart_arr[1]} ...try ${try}/${maxtry}"
      helm chart pull "${HELM_REGISTRY_URL}"/"${crt_chart_arr[0]}":"${crt_chart_arr[1]}" &>> "${dev_null}" && status="ready"

      [[ ${status} == "ready" ]] || sleep 2
    done
    [[ ${status} == "ready" ]] || error_with_code "10022" "${HELM_REGISTRY_URL}/${crt_chart_arr[0]}:${crt_chart_arr[1]}"
    helm chart export "${HELM_REGISTRY_URL}"/"${crt_chart_arr[0]}":"${crt_chart_arr[1]}" --destination "${TMP_HELM_CHARTS_DIR}" &>> "${dev_null}"
  done
}

function import_offline_infra_artifacts() {
  local rke2_bundle="${OFFLINE_INSTALL_FOLDER}/rke2-bundle.tar.gz"
  local rke2_sha="${OFFLINE_INSTALL_FOLDER}/rke2-bundle-sha256sum.txt"
  local infra_bundle="${OFFLINE_INSTALL_FOLDER}/infra-bundle.tar.gz"
  local infra_sha="${OFFLINE_INSTALL_FOLDER}/infra-bundle-sha256sum.txt"
  local offline_rke2_dir="${OFFLINE_INSTALL_FOLDER}/rke2"
  local rke2_agent_images_dir="/var/lib/rancher/rke2/agent/images"
  local offline_infra_dir="${OFFLINE_INSTALL_FOLDER}/infra"
  local infra_chart_inventory="${OFFLINE_INSTALL_FOLDER}/infra/helm/helm-tags.txt"
  local rke2_agent_temp_images_dir="${rke2_agent_images_dir}/temp_images"

  check_bundle "${rke2_bundle}" "${rke2_sha}"
  info "Extracting ${rke2_bundle} to ${offline_rke2_dir}"
  mkdir -p "${offline_rke2_dir}" && tar -xzf "${rke2_bundle}" -C "${offline_rke2_dir}"
  mkdir -p "${rke2_agent_images_dir}"
  /bin/mv -f "${offline_rke2_dir}"/*.tar.zst "${rke2_agent_images_dir}"
  rm -f "${rke2_bundle}"

  export INSTALL_RKE2_ARTIFACT_PATH="${offline_rke2_dir}"
  info "Set INSTALL_RKE2_ARTIFACT_PATH to ${offline_rke2_dir}"

  check_bundle "${infra_bundle}" "${infra_sha}"
  info "Extracting ${infra_bundle} to ${offline_infra_dir}"
  mkdir -p "${rke2_agent_temp_images_dir}" && tar -xzf "${infra_bundle}" -C "${rke2_agent_temp_images_dir}" ./images
  mkdir -p "${offline_infra_dir}" && tar -xzf "${infra_bundle}" -C "${offline_infra_dir}" ./helm
  rm -f "${infra_bundle}"

  cp -r "${rke2_agent_temp_images_dir}/images"/* "${rke2_agent_images_dir}"
  info "Infra images copied to ${rke2_agent_images_dir}"
  info "Cleaning up ${rke2_agent_temp_images_dir}"
  rm -rf "${rke2_agent_temp_images_dir}"

  info "Preparing the infra helm charts folder"
  [[ -f "${infra_chart_inventory}" ]] || error_with_code "10006" "${infra_chart_inventory}"
  TMP_HELM_CHARTS_DIR="${offline_infra_dir}/helm"
  while read -r line; do
    local chart_name file_name
    chart_name=$(awk '{print $1}' <<< "${line}")
    file_name=$(awk '{print $2}' <<< "${line}")
    info "Expanding ${file_name} to ${TMP_HELM_CHARTS_DIR}/${chart_name}"
    helm chart save "${TMP_HELM_CHARTS_DIR}/${file_name}" "localhost:${DOCKER_REGISTRY_NODEPORT}/${chart_name}" >> "${dev_null}"
    helm chart export "localhost:${DOCKER_REGISTRY_NODEPORT}/${chart_name}" --destination "${TMP_HELM_CHARTS_DIR}" >> "${dev_null}"
  done < "${infra_chart_inventory}"
}

function setup_offline_infra() {
  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    import_offline_infra_artifacts
  else
    TMP_HELM_CHARTS_DIR=$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t sf-infra-helm-charts.XXXXXXXXXX)
    download_infra_charts
  fi
  info "Using infra helm charts from ${TMP_HELM_CHARTS_DIR}"
  export TMP_HELM_CHARTS_DIR
}

function install_infra() {
  setup_offline_infra
  "${SCRIPT_DIR}"/Infra_Installer/infra-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/infra_output.json" "${DEBUG_FLAG[@]}"
  info "Cleaning up temporary folder ${OFFLINE_INSTALL_FOLDER}/rke2"
  rm -rf "${OFFLINE_INSTALL_FOLDER}/rke2"
}

function install_fabric() {

  "${SCRIPT_DIR}"/Fabric_Installer/fabric-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/fabric_output.json" "${DEBUG_FLAG[@]}"
}

function install_services() {
  "${SCRIPT_DIR}"/Services_Installer/services-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/services_output.json" "${DEBUG_FLAG[@]}"
}

function do_upgrade() {
  (( "${INSTALLINFRA}" == "1" )) && setup_offline_infra
  "${SCRIPT_DIR}"/Upgrade/upgrade.sh -i "${INFILE}" -o "${SCRIPT_DIR}/upgrade_output.json" -c "${INSTALLCNI}" -k "${INSTALLINFRA}" -f "${INSTALLFABRIC}" -s "${INSTALLSERVICES}" -a "${INSTALLSTANDALONEAICENTER}" "${DEBUG_FLAG[@]}"
}

function install_aicenter_services() {
  "${SCRIPT_DIR}"/Services_Installer/aicenter-service-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/services_output.json" "${DEBUG_FLAG[@]}"
}

function setup_backup() {
    /bin/bash "${DEBUG_ARGS[@]}" "${SCRIPT_DIR}"/Infra_Installer/disaster_recovery/backup.sh -i "${INFILE}" -o "${SCRIPT_DIR}/backup_output.json"
}

function setup_restore() {
    if [[ -z "${RESTORE_VOLUME}" ]]; then
      setup_offline_infra
    fi
    /bin/bash "${DEBUG_ARGS[@]}" "${SCRIPT_DIR}"/Infra_Installer/disaster_recovery/restore.sh -i "${INFILE}" -o "${SCRIPT_DIR}/restore_output.json"
}

function check_eula() {
    [[ -z "${LICENSE_AGREEMENT}" ]] || EULA="${LICENSE_AGREEMENT}"
    if [[ "${EULA}" != "accept" ]]; then
      echo "Use of paid UiPath products and services is subject to the licensing agreement executed between you and UiPath."
      echo "Unless otherwise indicated by UiPath, use of free UiPath products is subject to the associated licensing agreement available here: https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website)."
      echo "By setting the environment variable LICENSE_AGREEMENT=accept or by using the --accept-license-agreement argument, you agree to the applicable licensing agreement."
      exit 0
    fi
}

function validate_cluster_requirement() {
  local allow_installation_value
  allow_installation_value=$("${MODULES_DIR}"/cluster-validation.sh)

  echo "${allow_installation_value}"
}

function set_offline_variables() {
  local docker_registry_password_name="docker-registry-credentials"
  HELM_REGISTRY_URL="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
  HELM_REGISTRY_PASSWORD=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.password\}\} | base64 -d)
  HELM_REGISTRY_USERNAME=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.username\}\} | base64 -d)
}

function parse_input(){
    export ORIGINAL_INFILE="${INFILE}"
    merge_default_values && export INFILE
    parse_input_file
    info "Provided input file: ${INFILE}"
}

function read_certificate_files() {
  if [[ (-z "${NODE_ROLE}" || "${NODE_ROLE}" == "none") && "${ETCD_RESTORE}" != "true" ]]; then
    parse_certificate_file_paths

    if [[ ! -f "${CERT_CA_FILE_PATH}" || ! -f "${CERT_TLS_FILE_PATH}" || ! -f "${CERT_KEY_FILE_PATH}" ]]; then
      info "One or more certificate files (main FQDN) not found..."
      info "Generating TLS certificate..."
      create_self_issued_certificate_basic "${CLUSTER_FQDN}" "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}"

      # adding certificate in the trust store of node
      cp "${CERT_CA_FILE_PATH}" "/etc/pki/ca-trust/source/anchors/"
      update-ca-trust
    fi
    if [[ (-n "${ALTERNATIVE_FQDN}") && (! -f "${ALT_CERT_CA_FILE_PATH}" || ! -f "${ALT_CERT_TLS_FILE_PATH}" || ! -f "${ALT_CERT_KEY_FILE_PATH}") ]]; then
      info "One or more certificate files (alternative) not found..."
      info "Generating TLS certificate..."

      create_self_issued_certificate_basic "${ALTERNATIVE_FQDN}" "${ALT_CERT_CA_FILE_PATH}" "${ALT_CERT_TLS_FILE_PATH}" "${ALT_CERT_KEY_FILE_PATH}"

      # adding certificate in the trust store of node
      cp "${ALT_CERT_CA_FILE_PATH}" "/etc/pki/ca-trust/source/anchors/"
      update-ca-trust
    fi

    # Initializing to success as this is the case for standalone. For active_active it will be reset by the
    # validate_server_certificate function return value.
    alt_cert_validation_status="success"
    if [[ "${INSTALLSTANDALONEAICENTER}" == "0" ]]; then
      cert_validation_status=$(validate_server_certificate "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${CLUSTER_FQDN}" "${ARGO_SUBDOMAIN}" "${OBJECTSTORE_SUBDOMAIN}" "${DOCKER_REGISTRY_SUBDOMAIN}" "${MONITORING_SUBDOMAIN}" "${INSIGHTS_SUBDOMAIN}")
      if [[ "${ACTIVE_ACTIVE:-false}" == "true" ]]; then
        alt_cert_validation_status=$(validate_server_certificate "${ALT_CERT_CA_FILE_PATH}" "${ALT_CERT_TLS_FILE_PATH}" "${ALT_CERT_KEY_FILE_PATH}" "${ALTERNATIVE_FQDN}" "${ARGO_SUBDOMAIN}" "${OBJECTSTORE_SUBDOMAIN}" "${DOCKER_REGISTRY_SUBDOMAIN}" "${MONITORING_SUBDOMAIN}" "${INSIGHTS_SUBDOMAIN}")
      fi
    else
      cert_validation_status=$(validate_server_certificate "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${CLUSTER_FQDN}" "${ARGO_SUBDOMAIN}" "${OBJECTSTORE_SUBDOMAIN}" "${DOCKER_REGISTRY_SUBDOMAIN}" "${MONITORING_SUBDOMAIN}")
    fi
    [[ "${cert_validation_status}" == "success" && "${alt_cert_validation_status}" == "success" ]] && info "Validating Server Certificate... Done"

    update_fqdn=$(get_cluster_config "update_fqdn")
    if [[ ("${cert_validation_status}" != "success" || "${alt_cert_validation_status}" != "success") && "${update_fqdn}" == "true" ]]; then
      if [[ "${FORCEUPDATEFQDN}" == "0" ]]; then
        cat << EOF

Certificate provided is invalid for new FQDN, would you like us to configure new self signed certificate?
If you wish to continue, type 'yes' and hit Enter for confirmation.

EOF
        read -r force </dev/tty
        if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]; then
          info "Proceeding to configure new certificate"
        else
          info "Unexpected user input received: ${force}, exiting cluster fqdn update operation"
          exit 2
        fi
      else
        info "Proceeding to configure new certificate with force option"
      fi
      create_self_issued_certificate_basic "${CLUSTER_FQDN}" "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}"

      # adding certificate in the trust store of node
      cp "${CERT_CA_FILE_PATH}" "/etc/pki/ca-trust/source/anchors/"

      if [[ -n "${ALTERNATIVE_FQDN}" ]]; then
        create_self_issued_certificate_basic "${ALTERNATIVE_FQDN}" "${ALT_CERT_CA_FILE_PATH}" "${ALT_CERT_TLS_FILE_PATH}" "${ALT_CERT_KEY_FILE_PATH}"
        cp "${ALT_CERT_CA_FILE_PATH}" "/etc/pki/ca-trust/source/anchors/"
      fi
      update-ca-trust
    fi

    export CERT_CA_FILE_PATH
    export CERT_TLS_FILE_PATH
    export CERT_KEY_FILE_PATH
    export ALT_CERT_CA_FILE_PATH
    export ALT_CERT_TLS_FILE_PATH
    export ALT_CERT_KEY_FILE_PATH

    if [[ ${INSTALLSTANDALONEAICENTER} == 0 ]]; then
      parse_token_signing_certs_path

      if [[ ! -f "${TOKEN_SIGNING_CERT}" || -z "${TOKEN_SIGNING_PASS}" ]]; then
        info "Token signing certificate file or corresponding password not found..."
        info "Generating token signing certificate..."
        local reset_pipefail=false
        [[ -o pipefail ]] && reset_pipefail=true
        set +o pipefail
        LC_CTYPE=C TOKEN_SIGNING_PASS="$(tr </dev/urandom -dc _A-Z-a-z-0-9 | head -c 16)"
        [[ ${reset_pipefail} == true ]] && set -o pipefail
        create_token_signing_certificate "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${TOKEN_SIGNING_CERT}" "${TOKEN_SIGNING_PASS}"
      fi
      token_signing_cert_validation=$(validate_signing_certificate "${TOKEN_SIGNING_CERT}" "${TOKEN_SIGNING_PASS}")

      if [[ "${token_signing_cert_validation}" == "success" ]];
      then
        info "Validating Token Signing Certificate..... DONE"
      else
        error "Validating Token Signing Certificate..... FAILED"
      fi

      export TOKEN_SIGNING_CERT
      export TOKEN_SIGNING_PASS
    fi
  fi
}

# Checks if previous install type is in sync with the current one, offline or online
function verify_install_type() {
  local existing_install_type
  if systemctl is-enabled rke2-server &>>"${dev_null}" || systemctl is-enabled rke2-agent &>>"${dev_null}";
  then
    if rpm -q rke2-common &>>"${dev_null}";
    then
      existing_install_type="online"
    else
      existing_install_type="offline"
    fi

    # Error out if the existing install type does not match the current one
    if [[ "${existing_install_type}" != "${1}" ]];
    then
      error_with_code "10024" "${existing_install_type}"
    fi
  fi
}

function is_cluster_installed() {
  [[ "$(systemctl is-enabled rke2-server 2>>"${dev_null}")" == "enabled" ]] && echo -n "true" && return
  [[ "$(systemctl is-enabled rke2-agent 2>>"${dev_null}")" == "enabled" ]] && echo -n "true" && return
  echo "false"
}


function is_updating_fqdn() {
  local current_cluster_fqdn
  local alternative_cluster_fqdn
  if [[ $(is_cluster_installed) == "true" ]];
  then
    if [[ "${ACTIVE_ACTIVE:-false}" == "true" ]]; then
      alternative_cluster_fqdn=$(get_cluster_config "alternative_fqdn")
    fi
    # Check if alternative_cluster_fqdn is empty as this will be the case prior to 22.4.1 and if so
    # retrieve cluster_fqdn key value from the secret for comparison.
    if [[ -n "${alternative_cluster_fqdn}" ]]; then
      current_cluster_fqdn="${alternative_cluster_fqdn}"
    else
      current_cluster_fqdn=$(get_cluster_config "cluster_fqdn")
    fi
    update_fqdn=$(get_cluster_config "update_fqdn")
    {
      # Post 22.4.1 ALTERNATIVE_FQDN is set to CLUSTER_FQDN if its empty or else will use the value
      # provided in the input json file.
      [[ ${update_fqdn} == "true" ]] || [[ -n "${current_cluster_fqdn}" && "${current_cluster_fqdn}" != "${ALTERNATIVE_FQDN}" ]];
    } && echo -n "true" && return
  fi
  echo -n "false"
}

function check_and_warn_fqdn_update() {
  if [[ $(is_updating_fqdn) == "true" ]]; then

    if [[ "${FORCEUPDATEFQDN}" == "0" ]]; then
  cat << EOF


Changing Cluster FQDN may result in disruption of Automation Suite usage.
It may disconnect your Robots, Mobile Orchestrator users, ML Activities, ML Skills,
and it may invalidate any pending confirmation emails.

If you wish to continue, type 'yes' and hit Enter for confirmation.
EOF
      read -r force </dev/tty
      if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]; then
        info "Proceeding to update cluster fqdn"
      else
        info "Unexpected user input received: ${force}, exiting cluster fqdn update operation"
        exit 2
      fi
    else
      info "Proceeding to update cluster fqdn with force option"
    fi

    local service_install_type
    service_install_type=$(get_cluster_config "SERVICE_INSTALL_TYPE")
    if [[ "${service_install_type}" == "aicenter" ]]; then
      export INSTALLSTANDALONEAICENTER=1
    fi
    update_cluster_config "update_fqdn" "true"
    export UPDATE_FQDN=1
  fi
}

function check_and_warn_config_update() {
  local current_config_path user_input certificate_fields jq_delpaths_cmd

  certificate_fields='[["identity_certificate"],["server_certificate"]]'
  jq_delpaths_cmd="jq --sort-keys 'delpaths($certificate_fields)'" # using eval because passing anything but strings to jq args is a nightmare

  kubectl get application -n argocd uipath >> "${dev_null}" || return 0 # check if this is the first install

  current_config_path=$(mktemp -p "${DOT_FOLDER}" -t cluster_config.XXXXXXXXXX.json)

  echo "Found an existing cluster. Checking for config differences... " >&3
  # get the old settings
  (INFILE="${DEFAULTSFILE}" "${SCRIPT_DIR}"/configureUiPathAS.sh config "get" -o "${current_config_path}" "${DEBUG_FLAG[@]}" &>> "${dev_null}") || error_with_code "12015"

  # compare old settings with new settings ignoring certificate paths (those are handled by the config CLI)
  diff -q <(eval "${jq_delpaths_cmd}" < "${current_config_path}") <(jq -s '.[0] * .[1]' "${DEFAULTSFILE}" "${ORIGINAL_INFILE}" | eval "${jq_delpaths_cmd}") >> "${dev_null}" \
    || {
        echo "The config you are trying to apply will change the following settings:" >&3
        diff -u --color=always --label old --label new \
            <(eval "${jq_delpaths_cmd}" < "${current_config_path}") \
            <(jq -s '.[0] * .[1]' "${DEFAULTSFILE}" "${ORIGINAL_INFILE}" | eval "${jq_delpaths_cmd}") >&3 || true

        echo -e "\e[0;33mSome changes might result in downtime. For more details check:\e[0m" >&3
        echo -e "\e[0;33mhttps://docs.uipath.com/automation-suite/docs/advanced-installation-experience#general-configuration\e[0m" >&3
        echo "Do you wish to continue with the new settings? [y/n]" >&3
        while :;do
          echo -n "Type your choice and press Enter: " >&3
          read -r user_input </dev/tty
          if printf "%s\n" "${user_input}" | grep -Eq "$(locale yesexpr)"
            then
              echo "Continuing the install process using the new values..." >&3
              break
            elif printf "%s\n" "$user_input" | grep -Eq "$(locale noexpr)"
            then
              echo "Aborting the installer..." >&3
              exit 2
            else
              echo "Invalid choice" >&3
          fi
        done
    }

  return 0
}

function get_agent_type() {
  local taints
  taints=$(< "/etc/rancher/rke2/config.yaml" grep "node-taint")
  local agent_type="agent"
  if [[ -n "${taints}" ]]; then
    if [[ "${taints}" == *"task.mining"* ]]; then
      agent_type="task-mining"
    elif [[ "${taints}" == *"gpu"* ]]; then
      agent_type="gpu"
    fi
  fi
  echo "${agent_type}"
}

function is_tls_cert_present_in_cluster() {
  CHECKRERUN=0
  local namespace="istio-system"
  local secret_name="istio-ingressgateway-certs"
  if kubectl get secret ${secret_name} --namespace ${namespace}  >> /dev/null 2>&1; then
    info "This is the Re-run of the Installation, will be skipping the pre-check for the cert"
    CHECKRERUN=1
  fi
}

function check_prerequisites() {
  local offlineInstallFolder="$1"

  local prerequisites_args=()
  prerequisites_args+=(--install-type "${INSTALL_TYPE}")
  prerequisites_args+=(--input "${INFILE}")

  [[ -n "${MACHINES}" ]] && prerequisites_args+=(--machines "${MACHINES}")
  [[ ${INSTALLSTANDALONEAICENTER} -eq 1 ]] && prerequisites_args+=(--install-standalone-aicenter)
  if [[ ${NODE_TYPE} == "none" ]]; then
    prerequisites_args+=(--primary-server)
    prerequisites_args+=(--node-type "server")
  else
    prerequisites_args+=(--node-type "${NODE_TYPE}")
  fi

  if [[ "${INSTALL_TYPE}" == "offline" ]]; then
    if [[ -n "${OFFLINE_BUNDLE_PATH}" ]]; then
      prerequisites_args+=(--prereq-bundle "${OFFLINE_BUNDLE_PATH}")
      prerequisites_args+=(--prereq-tmp-folder "${offlineInstallFolder}")
    else
      prerequisites_args+=(--skip-bundle-validation)
    fi
    prerequisites_args+=(--skip-bundle-extraction)
  fi

  if [[ $(is_updating_fqdn) == "true" ]]; then
    prerequisites_args+=(--disabled-prereq-check "certificates")
  fi

  if [[ "${CHECKRERUN}" != "0" ]]; then
    prerequisites_args+=(--disabled-prereq-check "certificates")
  fi

  if [[ "${SETUPRESTORE}" == "1" ]]; then
    prerequisites_args+=(--disabled-prereq-check "sql,proxy,connectivity,certificates")
  fi

  if [[ "${DO_UPGRADE}" == "1" ]]; then
    prerequisites_args+=(--disabled-prereq-check "certificates")
  fi
  /bin/bash "${DEBUG_ARGS[@]}" "${SCRIPT_DIR}/validateUiPathASReadiness.sh" "${prerequisites_args[@]}"
}


function send_upgrade_telemetry() {
  local telemetry_content
  local current_version

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  ensure_machine_hash

  [[ ${NODE_ROLE} == "agent" ]] && export KUBECONFIG="/var/lib/rancher/rke2/agent/kubelet.kubeconfig"

  if [[ ${NODE_ROLE} == "none" ]]; then
    current_version=$(get_best_guess_current_version "server")
  else
    current_version=$(get_best_guess_current_version "${NODE_ROLE}")
  fi

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Upgrade.Input",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "BaseVersion": "${current_version:-none}",
                "NodeRole": "${NODE_ROLE:-none}",
                "ZeroDowntime" : "${ZDS:-none}",
                "ToVersion" : "${VERSION:-none}",
                "InstallId": "${INSTALL_ID:-none}",
                "ClusterToken" : "${RKE_TOKEN}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function main() {

  local allow_installation_of_services="true"
  if (( SETUPBACKUP == 0 )); then
    parse_input
  fi

  check_eula
  load_os_utils "${MODULES_DIR}"
  export KUBECONFIG="${KUBECONFIG_PATH:-/etc/rancher/rke2/rke2.yaml}"

  if (( SETUPRESTORE + SETUPBACKUP + DO_UPGRADE == 0 )); then
    (( COMPARE_CONFIG == 0 )) || { check_and_warn_config_update; check_result="$?";}
    (( COMPARE_CONFIG_ONLY == 0 )) || exit ${check_result}
  fi

  #checking if this rereun or fresh installation
  is_tls_cert_present_in_cluster

  # If upgrading, determine current node type.
  if [[ "${DO_UPGRADE}" == "1" ]]; then
    [[ "$(systemctl is-enabled rke2-server 2>>"${dev_null}")" == "enabled" ]] && NODE_ROLE="server" && NODE_TYPE="server"
    [[ "$(systemctl is-enabled rke2-agent 2>>"${dev_null}")" == "enabled" ]] && NODE_ROLE="agent" && NODE_TYPE=$(get_agent_type)
    [[ -z ${NODE_ROLE} ]] && error_with_code "11028"
    if [[ ${NODE_ROLE} == "server" && $(determine_first_node_upgrade "${VERSION}") == "true" ]]; then
      NODE_ROLE="none"
      CURRENT_VERSION="$(kubectl get applications/uipath -n argocd -o jsonpath='{.spec.source.targetRevision}')"
      for node in $(kubectl get nodes -o name);
      do
        add_resource_annotation_value "${node}" "uipath/infra-version" "${CURRENT_VERSION}"
      done
    fi
    [[ "${INSTALLCNI}" == "0" ]] && send_upgrade_telemetry
  fi

  [[ ${NODE_ROLE} == "agent" ]] && export KUBECONFIG="/var/lib/rancher/rke2/agent/kubelet.kubeconfig"
  export INSTALLSTANDALONEAICENTER
  export NODE_ROLE=${NODE_ROLE:-none}
  export NODE_TYPE=${NODE_TYPE:-none}

  [[ "${INSTALLCNI}" == "1" || "${INSTALLINFRA}" == "1" ]] && require_root_user

  [[ -d "${DOT_FOLDER}" ]] || mkdir -p "${DOT_FOLDER}"

  if [[ "${DO_UPGRADE}" == "1" && ("${INSTALLCNI}" == "1" || "${INSTALLINFRA}" == "1") && $(determine_first_node_upgrade "${VERSION}") == "true" ]];
  then
    cluster_backup
  fi

  if (( INSTALLINFRA + INSTALLFABRIC + INSTALLSERVICES + DO_UPGRADE + SETUPRESTORE != 0 )); then
    if [[ ${INSTALL_TYPE} == "offline" ]]; then
      if [[ -z "${RESTORE_VOLUME}" ]]; then
        verify_install_type "offline"
        load_prerequisites_package_utils "${MODULES_DIR}"
        if [[ -n ${OFFLINE_BUNDLE_PATH} ]]; then
          info "Running the offline install using ${OFFLINE_BUNDLE_PATH}"
          OFFLINE_INSTALL_FOLDER="${OFFLINE_TMP_FOLDER}/sf_installer"
          extract_offline_bundle "${OFFLINE_BUNDLE_PATH}" "${OFFLINE_INSTALL_FOLDER}"
          info "Files used for the offline installer have been extracted to ${OFFLINE_INSTALL_FOLDER}"
          if (( INSTALL_OFFLINE_PREREQS != 0 )); then
            install_offline_prereqs "${OFFLINE_INSTALL_FOLDER}"
          fi
        fi
        [[ "${SKIP_PRE_REQS}" == "0" ]] && check_prerequisites "${OFFLINE_INSTALL_FOLDER}"
      fi
    else
      verify_install_type "online"
      [[ "${SKIP_PACKAGES_INSTALL}" == "0" ]] && install_package_dependencies
      [[ "${SKIP_PRE_REQS}" == "0" ]] && check_prerequisites
      OFFLINE_INSTALL_FOLDER=""
      info "Running the online install"
    fi
  fi

  if (( SETUPRESTORE + SETUPBACKUP == 0 )); then
    check_and_warn_fqdn_update
  fi

  info "###############################################"
  info ""
  info "   Running UiPath Automation Suite Installer"
  info "             Version ${VERSION}"
  info ""
  info "###############################################"


  export INSTALL_TYPE OFFLINE_INSTALL_FOLDER

  if (( SETUPRESTORE + SETUPBACKUP + DO_UPGRADE == 0 )); then
    if [[ "${CHECKRERUN}" == "0" || $(is_updating_fqdn) == "true" ]]; then
      read_certificate_files
    fi
  fi

  [[ -n ${NODE_ROLE} ]] && export NODE_ROLE=${NODE_ROLE}
  [[ -n ${NODE_TYPE} ]] && export NODE_TYPE=${NODE_TYPE}

  [[ -n "${RESTORE_VOLUME}" ]] && export RESTORE_VOLUME=${RESTORE_VOLUME}
  [[ -n "${DISABLE_BACKUP}" ]] && export DISABLE_BACKUP=${DISABLE_BACKUP}
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLINFRA}" == "1" ]] && install_infra
  [[ -n "${ONE_TIME_MANUAL_BACKUP}" ]] && export ONE_TIME_MANUAL_BACKUP=${ONE_TIME_MANUAL_BACKUP}
  # get secrets access is not available in agent nodes kubeconfig so disabling this for agent node
  if [[ ${INSTALL_TYPE} == "offline" && "${NODE_ROLE}" != "agent" && "${SETUPRESTORE}" != "1" ]]; then
    set_offline_variables
  fi
  [[ -n ${OFFLINE_BUNDLE_PATH} ]] && [[ "${INSTALLFABRIC}" == "1" ]] && import_offline_sf_artifacts "${PODMAN_FOLDER}" "${OFFLINE_INSTALL_FOLDER}"
  [[ "${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}|${INSTALLSTANDALONEAICENTER}" == "1|1|1|0" && "${INSTALL_MULTINODE}" == "true" ]] && allow_installation_of_services="$(validate_cluster_requirement)"
  [[ "${allow_installation_of_services}" == "false" ]] && warn "Skipping fabric and service installation because minimum cluster resource requirements are not met, Please check above warnings!"
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLFABRIC}" == "1" && "${allow_installation_of_services}" == "true" ]] && install_fabric
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLSERVICES}" == "1" && "${INSTALLSTANDALONEAICENTER}" == "0" && "${allow_installation_of_services}" == "true" ]] && install_services
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLSERVICES}" == "1" && "${INSTALLSTANDALONEAICENTER}" == "1" ]] && install_aicenter_services
  if [[ "${SETUPBACKUP}" == "1" ]];
  then
    setup_backup
    if [[ "${NODE_ROLE}" == "none" ]];
    then
      save_restore_json "${INFILE}"
    fi
  fi

  # Customer will use this namespace to deploy any custom resource like container security scanner workload
  if (( INSTALLINFRA + INSTALLFABRIC + INSTALLSERVICES + DO_UPGRADE != 0 )) && [[ ${NODE_ROLE} != "agent" ]]; then
      kubectl create namespace "security-system" --dry-run=client -o yaml | kubectl apply -f - &>>"${dev_null}"
      info "Created the security-system namespace"
  fi

  [[ "${SETUPRESTORE}" == "1" ]] && setup_restore
  [[ "${DO_UPGRADE}" == "1" ]] && do_upgrade

  echo "{}" >"${OUTFILE}"
  if [[ -f "${SCRIPT_DIR}/infra_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/infra_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/fabric_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/fabric_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/services_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/services_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/upgrade_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/upgrade_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi

  if (( DEBUG != 1)); then
    rm -rf "${OFFLINE_INSTALL_FOLDER}"
  fi

  # Close the custom file descriptor if opened by this script
  $FD3_OUTPUT || exec 3<&-
}

set -o pipefail
{
parse_long_args "$@"
main
} 2>&1 | tee -a "${INSTALLER_LOGFILE}"
